# RCAIDE/Library/Methods/Emissions/evaluate_correlation_emissions_indices.py
#  
# Created:  Jul 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import  RCAIDE
from    RCAIDE.Framework.Core import Data 
 
# package imports
import  numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  evaluate_correlation_emissions_indices
# ---------------------------------------------------------------------------------------------------------------------- 
def evaluate_correlation_emissions_indices(segment,settings,vehicle):
    """
    Computes emission indices using empirical correlations.

    Parameters
    ----------
    segment : Data
        Mission segment data container
        
        - state : Data
            Current state of the system

            - numerics : Data
                Numerical integration parameters

                - time : Data
                    Time integration settings
            - conditions : Data
                Flight conditions and component states
            - ones_row : function
                Creates array of ones with specified size
                
    settings : Data
        Configuration settings for the simulation
        
    vehicle : Data
        Vehicle configuration data

        - networks : list
            List of propulsion system networks

            - fuel_lines : list
                Fuel distribution systems

                - active : bool
                    Flag indicating if fuel line is in use
                - fuel_tanks : list
                    Fuel storage units

                    - tag : str
                        Identifier for the fuel tank
                    - fuel : Data
                        Fuel properties

                        - emission_indices : Data
                            Empirical emission indices

                            - NOx : float
                                NOx emission index [kg_NOx/kg_fuel]
                            - CO2 : float
                                CO2 emission index [kg_CO2/kg_fuel]
                            - H2O : float
                                H2O emission index [kg_H2O/kg_fuel]
                            - SO2 : float
                                SO2 emission index [kg_SO2/kg_fuel]
                            - Soot : float
                                Soot emission index [kg_soot/kg_fuel]
                        - global_warming_potential_100 : Data
                            100-year global warming potentials
            - propulsors : list
                Propulsion units (turbofans, turbojets, etc.)

    Returns
    -------
    None
        Updates segment.state.conditions.emissions with:
        
        - total : Data
            Total emissions over segment

            - NOx : float
                Total NOx emissions [kg]
            - CO2 : float
                Total CO2 emissions [kg]
            - H2O : float
                Total H2O emissions [kg]
            - SO2 : float
                Total SO2 emissions [kg]
            - Soot : float
                Total soot emissions [kg]
            - Contrails : float
                Total contrail effect [kg CO2 equivalent]
        - index : Data
            Emission indices

            - NOx : ndarray
                NOx emission index [kg_NOx/kg_fuel]
            - CO2 : ndarray
                CO2 emission index [kg_CO2/kg_fuel]
            - H2O : ndarray
                H2O emission index [kg_H2O/kg_fuel]
            - SO2 : ndarray
                SO2 emission index [kg_SO2/kg_fuel]
            - Soot : ndarray
                Soot emission index [kg_soot/kg_fuel]

    Notes
    -----
    This function uses pre-defined emission indices for each fuel type and integrates
    them over the mission segment based on fuel flow rates.

    **Major Assumptions**

    * Emission indices are constant for each fuel type
    * Indices are independent of operating conditions
    * Linear scaling with fuel flow rate

    **Theory**
    Total emissions are computed by:
    
    .. math::
        E_{i,total} = \int \dot{m}_{fuel}(t) \cdot EI_i \cdot GWP_i \, dt

    Where:

    * :math:`E_{i,total}` = Total emissions for species i
    * :math:`\dot{m}_{fuel}` = Fuel mass flow rate
    * :math:`EI_i` = Emission index for species i
    * :math:`GWP_i` = Global warming potential for species i
    
    Contrail effects are estimated by:
    
    .. math::
        E_{contrails} = \Delta R \cdot GWP_{contrails}

    Where:

    * :math:`\Delta R` = Flight range [km]
    * :math:`GWP_{contrails}` = Contrail global warming potential

    **Extra modules required**

    * numpy

    See Also
    --------
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method

    References
    ----------
    [1] Lee, D. S., et al. (2021). The contribution of global aviation to anthropogenic climate forcing for 2000 to 2018. Atmospheric Environment, 244, 117834.
    """  
    # unpack
    state      = segment.state
    I          = state.numerics.time.integrate
    NOx_total  = 0 * state.ones_row(1)  
    CO2_total  = 0 * state.ones_row(1) 
    SO2_total  = 0 * state.ones_row(1) 
    H2O_total  = 0 * state.ones_row(1) 
    Soot_total = 0 * state.ones_row(1) 

    for network in vehicle.networks:  
        for fuel_line in network.fuel_lines:
            if fuel_line.active: 
                for fuel_tank in fuel_line.fuel_tanks:
                    mdot = 0. * state.ones_row(1)   
                    for propulsor in network.propulsors:
                        for source in (propulsor.active_fuel_tanks):
                            if fuel_tank.tag == source:  
                                propulsor_results =  state.conditions.energy[propulsor.tag]
                                fuel =  fuel_tank.fuel
                                if (type(propulsor) ==  RCAIDE.Library.Components.Propulsors.Turbofan) or \
                                    type(propulsor) ==  RCAIDE.Library.Components.Propulsors.Turboprop or \
                                    type(propulsor) ==  RCAIDE.Library.Components.Propulsors.Turboshaft or \
                                    type(propulsor) ==  RCAIDE.Library.Components.Propulsors.Turbojet:    
                         
                                    EI_NOx  = fuel.emission_indices.NOx
                                    EI_CO2  = fuel.emission_indices.CO2 
                                    EI_H2O  = fuel.emission_indices.H2O
                                    EI_SO2  = fuel.emission_indices.SO2
                                    EI_Soot = fuel.emission_indices.Soot  
                                    
                                    mdot = propulsor_results.fuel_flow_rate
                                     
                                    # Integrate them over the entire segment
                                    NOx_total  += np.dot(I,mdot*EI_NOx)
                                    CO2_total  += np.dot(I,mdot*EI_CO2)
                                    SO2_total  += np.dot(I,mdot*EI_SO2)
                                    H2O_total  += np.dot(I,mdot*EI_H2O) 
                                    Soot_total += np.dot(I,mdot*EI_Soot)
                                     
         
    flight_range    =  state.conditions.frames.inertial.aircraft_range 
    Contrails_total =  (flight_range -   flight_range[0]) /1000 * fuel.global_warming_potential_100.Contrails

    emissions                 = Data()
    emissions.total           = Data()
    emissions.index           = Data() 
    emissions.total.NOx       = NOx_total   * fuel.global_warming_potential_100.NOx 
    emissions.total.CO2       = CO2_total   * fuel.global_warming_potential_100.CO2
    emissions.total.H2O       = H2O_total   * fuel.global_warming_potential_100.H2O  
    emissions.total.SO2       = SO2_total   * fuel.global_warming_potential_100.SO2  
    emissions.total.Soot      = Soot_total  * fuel.global_warming_potential_100.Soot 
    emissions.total.Contrails = Contrails_total   
    emissions.index.NOx       = EI_NOx   * state.ones_row(1)
    emissions.index.CO2       = EI_CO2   * state.ones_row(1)
    emissions.index.H2O       = EI_H2O   * state.ones_row(1)
    emissions.index.SO2       = EI_SO2   * state.ones_row(1)
    emissions.index.Soot      = EI_Soot  * state.ones_row(1)
    
    state.conditions.emissions =  emissions
    return   




# RCAIDE/Compoments/Wings/Control_Surfaces/Elevator.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports   
from .Control_Surface import Control_Surface 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Elevator
# ----------------------------------------------------------------------------------------------------------------------
class Elevator(Control_Surface):
    """
    A class representing an elevator control surface for aircraft pitch control.

    Attributes
    ----------
    tag : str
        Unique identifier for the elevator, defaults to 'elevator'
        
    hinge_fraction : float
        Location of the hinge line as fraction of chord, defaults to 0.0
        
    sign_duplicate : float
        Sign convention for duplicate elevator deflection, defaults to 1.0
        (synchronized deflection for pitch control)

    Notes
    -----
    The elevator is a trailing edge control surface used primarily for pitch control. 
    It inherits basic control surface functionality from the Control_Surface class 
    and adds specific attributes for elevator operation.
    
    **Definitions**

    'Hinge Fraction'
        The chordwise location of the elevator hinge line, measured from the leading 
        edge as a fraction of local chord
        
    'Sign Duplicate'
        Determines whether paired elevators deflect in the same or opposite directions. 
        1.0 indicates synchronized deflection for pitch control

    See Also
    --------
    RCAIDE.Library.Components.Wings.Control_Surfaces.Control_Surface
        Base class providing common control surface functionality
    RCAIDE.Library.Components.Wings.Control_Surfaces.Stabilator
        Alternative pitch control using all-moving horizontal tail
    """ 

    def __defaults__(self):
        """
        Sets default values for the elevator attributes.
        
        Notes
        -----
        See Control_Surface.__defaults__ for additional inherited attributes.
        """
        self.tag            = 'elevator'
        self.hinge_fraction = 0.0
        self.sign_duplicate = 1.0
        
        pass 


## @ingroup Core
# redirect.py
#
# Created:  Aug 2015, T. Lukacyzk
# Modified: Feb 2016, T. MacDonald

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

import os, sys, shutil, copy

# -------------------------------------------------------------------
#  Output Redirection 
# -------------------------------------------------------------------

## @ingroup Core
class output(object):
    """ Temporarily redirects sys.stdout and sys.stderr when used in
        a 'with' contextmanager
        
        Example:
        with SU2.io.redirect_output('stdout.txt','stderr.txt'):
            sys.stdout.write("standard out")
            sys.stderr.write("stanrard error")
            # code
        #: with output redirection
        
        Inputs:
            stdout - None, a filename, or a file stream
            stderr - None, a filename, or a file stream
        None will not redirect outptut
        
        Source:
        http://stackoverflow.com/questions/6796492/python-temporarily-redirect-stdout-stderr
        
    """
    def __init__(self, stdout=None, stderr=None):
        """ Initializes a new output() class
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """         
        
        _newout = False
        _newerr = False
        
        if isinstance(stdout,str):
            stdout = open(stdout,'a')
            _newout = True            
        if isinstance(stderr,str):
            stderr = open(stderr,'a')
            _newerr = True                   
                
        self._stdout = stdout or sys.stdout
        self._stderr = stderr or sys.stderr
        self._newout = _newout
        self._newerr = _newerr

    def __enter__(self):
        self.old_stdout, self.old_stderr = sys.stdout, sys.stderr
        self.old_stdout.flush(); self.old_stderr.flush()
        sys.stdout, sys.stderr = self._stdout, self._stderr

    def __exit__(self, exc_type, exc_value, traceback):
        self._stdout.flush(); self._stderr.flush()
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr
        
        if self._newout:
            self._stdout.close()
        if self._newerr:
            self._stderr.close()           


# -------------------------------------------------------------------
#  Folder Redirection 
# -------------------------------------------------------------------

## @ingroup Core
class folder(object):
    """ Temporarily redirects to a working folder, pulling 
        and pushing needed files
        
        Example:
        
        folder = 'temp'                    
        pull   = ['file1.txt','file2.txt'] 
        link   = ['file3.big']             
        force  = True                      
        
        # original path
        import os
        print os.getcwd()
        
        # enter folder
        with SU2.io.redirect_folder(folder,pull,link,force) as push:
            print os.getcwd()
            # code
            push.append('file4.txt')
        #: with folder redirection
        
        # returned to original path
        print os.getcwd()
        
        Inputs:
            folder - working folder, relative or absolute
            pull   - list of files to pull (copy to working folder)
            link   - list of files to link (symbolic link in working folder)
            force  - True/False overwrite existing files in working folder
        
        Targets:
            push   - list of files to push (copy to originating path)
        
        Notes:
            push must be appended or extended, not overwritten
            links in Windows not supported, will simply copy
    """
    
    def __init__(self, folder, pull=None, link=None, force=True ):
        ''' folder redirection initialization
            see help( folder ) for more info
        '''
        
        if pull is None: pull = []
        if link is None: link = []
        
        if not isinstance(pull,list) : pull = [pull]
        if not isinstance(link,list) : link = [link]
        
        origin = os.getcwd()
        origin = os.path.abspath(origin).rstrip('/')+'/'
        folder = os.path.abspath(folder).rstrip('/')+'/'
        
        self.origin = origin
        self.folder = folder
        self.pull   = copy.deepcopy(pull)
        self.push   = []
        self.link   = copy.deepcopy(link)
        self.force  = force

    def __enter__(self): 
        
        origin = self.origin  # absolute path
        folder = self.folder  # absolute path
        pull   = self.pull
        push   = self.push
        link   = self.link
        force  = self.force
        
        # check for no folder change
        if folder == origin:
            return []
        
        # check, make folder
        if not os.path.exists(folder):
            os.makedirs(folder)
        
        # copy pull files
        for name in pull:
            old_name = os.path.abspath(name)
            new_name = os.path.split(name)[-1]
            new_name = os.path.join(folder,new_name)
            if old_name == new_name: continue
            if os.path.exists( new_name ): 
                if force: os.remove( new_name )
                else: continue
            shutil.copy(old_name,new_name)

        # make links
        for name in link:
            old_name = os.path.abspath(name)
            new_name = os.path.split(name)[-1]
            new_name = os.path.join(folder,new_name)
            if old_name == new_name: continue
            if os.path.exists( new_name ): 
                if force: os.remove( new_name )
                else: continue
            make_link(old_name,new_name)
            
        # change directory
        os.chdir(folder)
        
        # return empty list to append with files to push to super folder
        return push

    def __exit__(self, exc_type, exc_value, traceback):
        
        origin = self.origin
        folder = self.folder
        push   = self.push
        force  = self.force
        
        # check for no folder change
        if folder == origin:
            return
        
        # move assets
        for name in push:
            
            old_name = os.path.abspath(name)
            name = os.path.split(name)[-1]
            new_name = os.path.join(origin,name)
            
            # links
            if os.path.islink(old_name):
                source = os.path.realpath(old_name)
                if source == new_name: continue
                if os.path.exists( new_name ):
                    if force: os.remove( new_name )
                    else: continue
                make_link(source,new_name)
            
            # moves
            else:
                if old_name == new_name: continue
                if os.path.exists( new_name ):
                    if force: os.remove( new_name )
                    else: continue
                shutil.move(old_name,new_name)
            
        # change directory
        os.chdir(origin)
        
  
def make_link(src,dst):
    """ make_link(src,dst)
        makes a relative link
        Inputs:
            src - source file
            dst - destination to place link
        
        Windows links currently unsupported, will copy file instead
    """
    
    assert os.path.exists(src) , 'source file does not exist \n%s' % src
    
    if os.name == 'nt':
        # can't make a link in windows, need to look for other options
        if os.path.exists(dst): os.remove(dst)
        shutil.copy(src,dst)
    
    else:
        # find real file, incase source itself is a link
        src = os.path.realpath(src) 
        
        # normalize paths
        src = os.path.normpath(src)
        dst = os.path.normpath(dst)        
        
        # check for self referencing
        if src == dst: return        
        
        # find relative folder path
        srcfolder = os.path.join( os.path.split(src)[0] ) + '/'
        dstfolder = os.path.join( os.path.split(dst)[0] ) + '/'
        srcfolder = os.path.relpath(srcfolder,dstfolder)
        src = os.path.join( srcfolder, os.path.split(src)[1] )
        
        # make unix link
        if os.path.exists(dst): os.remove(dst)
        os.symlink(src,dst)
    


# RCAIDE/Framework/Analyses/Mission/Segments/Climb/Constant_Dynamic_Pressure_Constant_Angle.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports  
from RCAIDE.Framework.Core                                     import Units 
from RCAIDE.Framework.Mission.Segments.Evaluate       import Evaluate
from RCAIDE.Library.Mission                          import Common,Segments

# ----------------------------------------------------------------------------------------------------------------------
# Constant_Dynamic_Pressure_Constant_Angle
# ---------------------------------------------------------------------------------------------------------------------- 
 
class Constant_Dynamic_Pressure_Constant_Angle(Evaluate):
    """ Climb at a constant dynamic pressure at a constant angle.This segment takes longer to solve than most because 
        it has extra unknowns and residuals
    
        Assumptions:
        None
        
        Source:
        None
    """       
    
    def __defaults__(self):
        """ This sets the default solver flow. Anything in here can be modified after initializing a segment.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """          
        
        # -------------------------------------------------------------------------------------------------------------- 
        #   User Inputs
        # -------------------------------------------------------------------------------------------------------------- 
        self.altitude_start            = None # Optional
        self.altitude_end              = 10.  * Units.km
        self.climb_angle               = 3.   * Units.degrees
        self.dynamic_pressure          = None
        self.true_course               = 0.0 * Units.degrees
        
        # -------------------------------------------------------------------------------------------------------------- 
        #  Mission Specific Unknowns and Residuals 
        # --------------------------------------------------------------------------------------------------------------    
        ones_row = self.state.ones_row             
        self.state.residuals.altitude      = ones_row(1) * 0.0
        self.state.unknowns.altitude       = ones_row(1) * 0.0                                         
        
        # -------------------------------------------------------------------------------------------------------------- 
        #  Mission specific processes 
        # --------------------------------------------------------------------------------------------------------------   
        initialize                         = self.process.initialize
        initialize.conditions              = Segments.Climb.Constant_Dynamic_Pressure_Constant_Angle.initialize_conditions_unpack_unknowns 
        iterate                            = self.process.iterate 
        iterate.unknowns.mission           = Common.Unpack_Unknowns.orientation
        iterate.unknowns.controls          = Common.Unpack_Unknowns.control_surfaces
        iterate.unknowns.kinematics        = Segments.Climb.Constant_Dynamic_Pressure_Constant_Angle.initialize_conditions_unpack_unknowns
        iterate.conditions.differentials   = Segments.Climb.Constant_Dynamic_Pressure_Constant_Angle.update_differentials 
        iterate.residuals.flight_dynamics  = Common.Residuals.flight_dynamics
        iterate.residuals.altitude         = Segments.Climb.Constant_Dynamic_Pressure_Constant_Angle.residual_altitude
        return
       


# RCAIDE/Library/Components/Propulsors/Electric_Ducted_Fan.py
# 
# 
# Created:  Oct 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports 
from .   import Propulsor  
from RCAIDE.Library.Methods.Propulsors.Electric_Ducted_Fan_Propulsor.append_electric_ducted_fan_conditions           import append_electric_ducted_fan_conditions
from RCAIDE.Library.Methods.Propulsors.Electric_Ducted_Fan_Propulsor.unpack_electric_ducted_fan_unknowns             import unpack_electric_ducted_fan_unknowns
from RCAIDE.Library.Methods.Propulsors.Electric_Ducted_Fan_Propulsor.pack_electric_ducted_fan_residuals              import pack_electric_ducted_fan_residuals 
from RCAIDE.Library.Methods.Propulsors.Electric_Ducted_Fan_Propulsor.compute_electric_ducted_fan_performance         import compute_electric_ducted_fan_performance, reuse_stored_electric_ducted_fan_data
from RCAIDE.Library.Methods.Propulsors.Electric_Ducted_Fan_Propulsor.append_electric_ducted_fan_residual_and_unknown import  append_electric_ducted_fan_residual_and_unknown

# ----------------------------------------------------------------------
#  Electric Ducted Fan Component
# ----------------------------------------------------------------------
class Electric_Ducted_Fan(Propulsor):
    """
    A propulsion system class that combines an electric motor with a ducted fan.
    
    Attributes
    ----------
    tag : str
        Identifier for the propulsion system, defaults to 'electric_ducted_fan'
    
    motor : None or Motor
        The electric motor component that provides rotational power
        
    ducted_fan : None or DuctedFan
        The ducted fan component that generates thrust
        
    electronic_speed_controller : None or ESC
        The electronic speed controller that regulates power to the motor
    
    Notes
    -----
    This class models an electric propulsion system where an electric motor drives 
    a ducted fan to generate thrust. The system includes an electronic speed controller 
    (ESC) to regulate power delivery from the electrical system to the motor.
    
    The ducted fan configuration provides several advantages over open propellers:
    - Higher static thrust efficiency
    - Reduced tip losses
    - Lower noise emission
    - Improved safety through shrouding
    
    **Definitions**

    'Ducted Fan'
        A propulsion device consisting of a fan enclosed within a duct/shroud 
        that improves thrust efficiency
    
    'Electronic Speed Controller (ESC)'
        Device that controls the speed of the electric motor by regulating 
        power delivery based on input commands
    
    See Also
    --------
    RCAIDE.Library.Components.Propulsors.Propulsor
    """
    def __defaults__(self):    
        # setting the default values
        self.tag                          = 'electric_ducted_fan'    
        self.motor                        = None
        self.ducted_fan                   = None 
        self.electronic_speed_controller  = None 

    def append_operating_conditions(self,segment):
        append_electric_ducted_fan_conditions(self,segment)
        return 

    def unpack_propulsor_unknowns(self,segment):  
        unpack_electric_ducted_fan_unknowns(self,segment)
        return 

    def pack_propulsor_residuals(self,segment): 
        pack_electric_ducted_fan_residuals(self,segment)
        return        

    def append_propulsor_unknowns_and_residuals(self,segment):
        append_electric_ducted_fan_residual_and_unknown(self,segment)
        return 
    
    def compute_performance(self,state,voltage,center_of_gravity = [[0, 0, 0]]):
        thrust,moment,power,stored_results_flag,stored_propulsor_tag =  compute_electric_ducted_fan_performance(self,state,voltage,center_of_gravity)
        return thrust,moment,power,stored_results_flag,stored_propulsor_tag
    
    def reuse_stored_data(EDF,state,bus,stored_propulsor_tag,center_of_gravity = [[0, 0, 0]]):
        thrust,moment,power = reuse_stored_electric_ducted_fan_data(EDF,state,bus,stored_propulsor_tag,center_of_gravity)
        return thrust,moment,power



# RCAIDE/Library/Components/Thermal_Management/Heat_Exchangers/Cryogenic_Heat_Exchange.py
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports  
import RCAIDE 
from RCAIDE.Library.Components import Component  
from RCAIDE.Library.Methods.Thermal_Management.Cryogenics.compute_cryogen_mass_flow_rate import compute_cryogen_mass_flow_rate
 
# ---------------------------------------------------------------------------------------------------------------------- 
#  Cryogenic Heat Exchanger Component
# ---------------------------------------------------------------------------------------------------------------------- 
class Cryogenic_Heat_Exchanger(Component):
    """
    A class representing a heat exchanger designed for cryogenic cooling applications, 
    using cryogenic fluids to maintain extremely low temperatures.

    Attributes
    ----------
    tag : str
        Unique identifier for the heat exchanger, defaults to 'Cryogenic_Heat_Exchanger'
        
    cryogen : Cryogen
        Cryogenic fluid properties, defaults to Liquid_H2()
        
    cryogen_inlet_temperature : float
        Temperature of cryogen at inlet in K, defaults to 300.0
        
    cryogen_outlet_temperature : float
        Temperature of cryogen at outlet in K, defaults to 300.0
        
    cryogen_pressure : float
        Operating pressure of cryogen in Pa, defaults to 100000.0
        
    cryogen_is_fuel : float
        Proportion of cryogen used as fuel, defaults to 0.0
        
    inputs.cooling_power : float
        Required cooling power in W, defaults to 0.0
        
    outputs.mdot : float
        Mass flow rate of cryogen in kg/s, defaults to 0.0

    Notes
    -----
    The cryogenic heat exchanger provides cooling at extremely low temperatures using 
    cryogenic fluids. The system:
    
    * Manages heat transfer between cryogen and cooled equipment
    * Controls cryogen flow based on cooling demand
    * Maintains equipment at cryogenic temperatures

    **Definitions**

    'Cryogen'
        Fluid used to achieve very low temperatures, typically below 120K
        
    'Cooling Power'
        Heat removal capacity at cryogenic temperature

    See Also
    --------
    RCAIDE.Library.Components.Thermal_Management.Heat_Exchangers.Cross_Flow_Heat_Exchanger
        Alternative heat exchanger for moderate temperature applications
    """
    
    def __defaults__(self):
        """
        Sets default values for the cryogenic heat exchanger attributes.
        """         
        self.tag                       = 'Cryogenic_Heat_Exchanger'
        self.cryogen                   = RCAIDE.Library.Attributes.Cryogens.Liquid_H2()
        self.cryogen_inlet_temperature = 300.0
        self.cryogen_outlet_temperature= 300.0
        self.cryogen_pressure         = 100000.0
        self.cryogen_is_fuel          = 0.0
        self.inputs.cooling_power     = 0.0
        self.outputs.mdot             = 0.0

    def energy_calc(self, conditions):
        """
        Calculates required cryogen mass flow rate for desired cooling power.

        Parameters
        ----------
        conditions : Data
            Operating conditions for the calculation

        Returns
        -------
        float
            Required cryogen mass flow rate in kg/s

        Notes
        -----
        Calculation considers:
        * Inlet and outlet temperatures
        * Required cooling power
        * Cryogen properties
        * System pressure
        """
        # unpack the values
        temp_in      = self.cryogen_inlet_temperature 
        temp_out     = self.cryogen_outlet_temperature
        pressure     = self.cryogen_pressure
        cryogen      = self.cryogen        
        cooling_power= self.inputs.cooling_power 
        vent_pressure= pressure

        # calculate the cryogen mass flow
        mdot = compute_cryogen_mass_flow_rate(cryogen, temp_in, temp_out, 
                                            cooling_power, vent_pressure)
        self.outputs.mdot = mdot
        return mdot
        
    


# RCAIDE/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Analysis  import Analysis 
from .Process   import Process
from .Settings  import Settings
from .Vehicle   import Vehicle 

from . import Common
from . import Aerodynamics
from . import Atmospheric
from . import Emissions
from . import Energy
from . import Noise
from . import Planets
from . import Propulsion
from . import Stability
from . import Weights
from . import Emissions
from . import Geodesics


# RCAIDE/Methods/Aerodynamics/Airfoil_Panel_Method/chordwise_distribution
# 
# 
# Created:  May 2024, Niranjan Nanjappa

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports    
from RCAIDE.Framework.Core import Data 

# package imports  
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
# chordwise_distribution
# ---------------------------------------------------------------------------------------------------------------------- 
def chordwise_distribution(x,y,cp,al,npanel,cf,vt):
    """Compute chordwise distribution of lift and drag for
        an airfoil          

    Assumptions:
    None

    Source:
    Anderson, J. D. "Fundamentals of aerodynamics 6th ed". MCGRAW-HILL EDUCATION (2017)                                                                  
                                                                   
    Inputs:                                                     
    x       -  Vector of x coordinates of the surface nodes      
    y       -  Vector of y coordinates of the surface nodes       
    cp      -  Vector of coefficients of pressure at the nodes  
    al      -  Angle of attack in radians                             
    npanel  -  Number of panels on the airfoil
    cf      -  skin friction coeff over the surface of airfoil
    vt      -  tangential velocity at all nodes   
                                                                 
    Outputs:                                           
    fD      -  chordwise distribution of drag
    fL      -  chordwise distribution of lift    
     
   Properties Used:
    N/A

    """
    al       = al.T
    dx       = x[1:]-x[:-1]
    dy       = y[1:]-y[:-1]
    LE_ind   = int(npanel/2)
    dx_l     = -np.flip(dx[0:LE_ind], axis=0)
    dx_u     = dx[LE_ind:]
    dy_l     = -np.flip(dy[0:LE_ind], axis=0)
    dy_u     = dy[LE_ind:]
    
    cp_l     = np.flip(cp[0:LE_ind], axis=0)
    cp_u     = cp[LE_ind:]
    
    cf       = cf*(vt/np.abs(vt))
    cf_l     = -np.flip(cf[0:LE_ind], axis=0)
    cf_u     = cf[LE_ind:]
    
    dcn      = cp_l*dx_l - cp_u*dx_u + cf_u*dy_u - cf_l*dy_l
    dca      = cp_u*dy_u - cp_l*dy_l + cf_u*dx_u + cf_l*dx_l
    
    dcl      = dcn*np.cos(al) - dca*np.sin(al)
    dcd      = dcn*np.sin(al) + dca*np.cos(al)
    
    cl       = np.sum(dcl, axis=0)
    cd       = np.sum(dcd, axis=0)
    
    fL       = dcl/cl
    fD       = dcd/cd
    
    return fL,fD
    


# RCAIDE/Library/Plots/Performance/Aerodynamics/plot_surface_pressures.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ---------------------------------------------------------------------------------------------------------------------- 
def plot_surface_pressures(results, save_figure = False, save_filename = "Surface_Pressure", file_type = ".png"):
    """This plots the surface pressure distrubtion at all control points
    on all lifting surfaces of the aircraft

    Assumptions:
    None

    Source:
    None

    Inputs:
    results.segments.aerodynamics.
        pressure_coefficient
    vehicle.vortex_distribution.
       n_cw
       n_sw
       n_w

    Outputs:
    Plots

    Properties Used:
    N/A
    """
    
    vehicle    = results.segments[0].analyses.aerodynamics.vehicle
    VD         = vehicle.vortex_distribution
    n_cw       = VD.n_cw
    n_cw       = VD.n_cw
    n_sw       = VD.n_sw
    n_w        = VD.n_w
    b_pts      = np.concatenate(([0],np.cumsum(VD.n_sw*VD.n_cw)))

    # Create a boolean for not plotting vertical wings
    idx        = 0
    plot_flag  = np.ones(n_w)
    for wing in vehicle.wings:
        if wing.vertical:
            plot_flag[idx] = 0
            idx += 1
        else:
            idx += 1
        if wing.vertical and wing.symmetric:
            plot_flag[idx] = 0
            idx += 1
        else:
            idx += 1

    img_idx    = 1
    seg_idx    = 1
    for segment in results.segments.values():
        num_ctrl_pts = len(segment.conditions.frames.inertial.time)
        for ti in range(num_ctrl_pts):
            CP         = segment.conditions.aerodynamics.coefficients.surface_pressure[ti]

            fig        = plt.figure()
            axes       = plt.subplot(1, 1, 1)
            x_max      = max(VD.XC) + 2
            y_max      = max(VD.YC) + 2
            axes.set_ylim(x_max, 0)
            axes.set_xlim(-y_max, y_max)
            fig.set_size_inches(8,8)
            for i in range(n_w):
                n_pts     = (n_sw[i] + 1) * (n_cw[i]+ 1)
                xc_pts    = VD.X[i*(n_pts):(i+1)*(n_pts)]
                x_pts     = np.reshape(np.atleast_2d(VD.XC[b_pts[i]:b_pts[i+1]]).T, (n_sw[i],-1))
                y_pts     = np.reshape(np.atleast_2d(VD.YC[b_pts[i]:b_pts[i+1]]).T, (n_sw[i],-1))
                z_pts     = np.reshape(np.atleast_2d(CP[b_pts[i]:b_pts[i+1]]).T, (n_sw[i],-1))
                x_pts_p   = x_pts*((n_cw[i]+1)/n_cw[i]) - x_pts[0,0]*((n_cw[i]+1)/n_cw[i])  +  xc_pts[0]
                points    = np.linspace(0.001,1,50)
                A         = np.cumsum(np.sin(np.pi/2*points))
                levels    = -(np.concatenate([-A[::-1],A[1:]])/(2*A[-1])  + A[-1]/(2*A[-1]) )[::-1]*0.015
                color_map = plt.cm.get_cmap('jet')
                rev_cm    = color_map.reversed() 
                CS        = axes.contourf(y_pts,x_pts_p, z_pts, cmap = rev_cm,extend='both')

            # Set Color bar
            cbar = fig.colorbar(CS, ax=axes)
            cbar.ax.set_ylabel('$C_{P}$', rotation =  0)
            plt.axis('off')
            plt.grid(None)

            if save_figure:
                plt.savefig( save_filename + '_' + str(img_idx) + file_type)
            img_idx += 1
        seg_idx +=1

    return



# RCAIDE/Library/Components/Energy/Sources/Fuel_Tanks/__init__.py
# 

"""
Fuel tank components for aircraft fuel storage and distribution

This module provides implementations for various aircraft fuel tank types including
wing tanks, central tanks, and generic fuel tanks. Each tank type has specific
characteristics for fuel storage, weight distribution, and system integration.

See Also
--------
RCAIDE.Library.Components.Energy.Distributors.Fuel_Line
    Fuel distribution components
RCAIDE.Library.Components.Energy.Modulators.Fuel_Selector
    Fuel flow control components
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Fuel_Tank  import Fuel_Tank
from .Central_Fuel_Tank import Central_Fuel_Tank
from .Wing_Fuel_Tank   import  Wing_Fuel_Tank


# RCAIDE/Library/Methods/Stability/Moment_of_Inertia/compute_fuselage_moment_of_inertia.py 
# 
# Created:  September 2024, A. Molloy  
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORTS
# ---------------------------------------------------------------------------------------------------------------------- 
# package imports 
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  Compute Fuselage Moment of Intertia
# ----------------------------------------------------------------------------------------------------------------------   
def compute_fuselage_moment_of_inertia(fuselage,center_of_gravity=[[0, 0, 0]]): 
    ''' computes the moment of ienrtia tensor for a generic fuselage about a given center of gravity. 

    Assumptions:
    - Fuselage can be approximated by a hemisphere, cylinder, and cone.
    - Fuselage is hollow with the inner radius being 85% of the outer radius.
    - Fuselage is constant density

    Source:
    [1] Moulton, B. C., and Hunsaker, D. F., “Simplified Mass and Inertial Estimates for Aircraft with Components
    of Constant Density,” AIAA SCITECH 2023 Forum, January 2023, AIAA-2023-2432 DOI: 10.2514/
    6.2023-2432
    
    [2] Weisstein, E. W., "Moment of Inertia -- Cone," Wolfram Research, N.D., https://scienceworld.wolfram.com/physics/MomentofInertiaCone.html 

    Inputs:
    - Fuselage
    - Center of gravity or the point to find the MOI about

    Outputs:
    - Fuselage moment of inertia tensor

    Properties Used:
    N/A
    '''
    
    # ----------------------------------------------------------------------------------------------------------------------   
    # Setup
    # ----------------------------------------------------------------------------------------------------------------------       
    mass_fuselage = 1 * fuselage.mass_properties.mass
    radius_outer  = fuselage.effective_diameter / 2
    radius_inner  = 0.85 * fuselage.effective_diameter / 2 # Assume the inner radius is 85 % of the outer radius    
    center_length = fuselage.lengths.total - fuselage.lengths.nose - fuselage.lengths.tail     # Length of the cylinder (found by subtracting the tail and nose lengths from the total length)
    
    I_total = np.zeros((3, 3)) # Initialize matrix to hold the entire fuselage inertia tensor
    
    # ---------------------------------------------------------------------------------------------------------------------- 
    # Calculate volume fraction of each section
    # ---------------------------------------------------------------------------------------------------------------------- 
    volume_fraction = Volume_Fraction(radius_outer, radius_inner, center_length, fuselage.lengths.tail) # output = [hemisphere, cylinder, cone]
    
    # ----------------------------------------------------------------------------------------------------------------------   
    # Hemisphere
    # ----------------------------------------------------------------------------------------------------------------------   
    origin_hemisphere = np.array([fuselage.lengths.nose, 0, 0]) + np.array(fuselage.origin)
    mass_hemisphere   = mass_fuselage * volume_fraction[0] # mass of the hemisphere
    I                 = np.zeros((3, 3)) # Local inertia tensor 
    
    # Moment of inertia in local system. From Weisstein [2]
    I[0][0] = 2 * mass_hemisphere / 5 *  (radius_outer ** 5 - radius_inner ** 5) /(radius_outer **3 -radius_inner **3) # Ixx
    I[1][1] = 2 * mass_hemisphere / 5 *  (radius_outer ** 5 - radius_inner ** 5) /(radius_outer **3 -radius_inner **3)# Iyy
    I[2][2] = 2 * mass_hemisphere / 5 *  (radius_outer ** 5 - radius_inner ** 5) /(radius_outer **3 -radius_inner **3) # Izz

    # global system
    s        = np.array(center_of_gravity) - np.array(origin_hemisphere)
    I_global = np.array(I) + mass_hemisphere * (np.array(np.dot(s[0], s[0])) * np.array(np.identity(3)) - s*np.transpose(s)) # global inertia tensor for hemisphere
    
    # Add hemisphere to the fuselage inertia tensor
    I_total  = np.array(I_total) + np.array(I_global)
    
    # ----------------------------------------------------------------------------------------------------------------------   
    # cylinder
    # ----------------------------------------------------------------------------------------------------------------------   
    mass_cylinder   = mass_fuselage * volume_fraction[1]
    origin_cylinder = np.array([fuselage.lengths.nose + center_length / 2,0, 0]) + np.array(fuselage.origin) # origin of the cylinder is located a the middle of the cylinder
    
    I =  np.zeros((3, 3))
    
    # Moment of inertia in local system. From Moulton and Hunsaker [1]
    I[0][0]  = mass_cylinder / 2 *  (radius_outer ** 2 + radius_inner ** 2) # Ixx
    I[1][1]  = mass_cylinder / 12 * (3 * (radius_outer ** 2 + radius_inner ** 2) + center_length ** 2) # Iyy
    I[2][2]  = mass_cylinder / 12 * (3 * (radius_outer ** 2 + radius_inner ** 2) + center_length ** 2) # Izz
      
    # transform moment of inertia to global system   
    s        = np.array(center_of_gravity) - np.array(origin_cylinder)
    I_global = np.array(I) + mass_cylinder * (np.array(np.dot(s[0], s[0])) * np.array(np.identity(3)) - s*np.transpose(s))
    
    # Add cylinder to fuselage inertia matrix
    I_total = np.array(I_total) + np.array(I_global)

    # ----------------------------------------------------------------------------------------------------------------------   
    # cone
    # ----------------------------------------------------------------------------------------------------------------------   
    tail_length = fuselage.lengths.tail # length of the cone is defined to be the tail length.
    mass_cone   = mass_fuselage * volume_fraction[2]
    origin_cone = np.array([fuselage.lengths.total - fuselage.lengths.tail,0, 0]) + np.array(fuselage.origin)
    
    I =  np.zeros((3, 3))
  
    # Moment of inertia in local system. From Weisstein [2]. 
    rho     = (mass_cone / (1 / 3 * np.pi * (radius_outer ** 2 * center_length - radius_inner ** 2 * (center_length * radius_inner / radius_outer)))) # density of the cone. Mass divided by volume.
    I[0][0] = rho * (1 / 3 * np.pi * radius_outer ** 2 * tail_length ** 3 + np.pi / 20 * radius_outer ** 4 *tail_length - 1 / 3 * np.pi * radius_inner ** 2 * (tail_length * radius_inner / radius_outer) ** 3 + np.pi / 20 * radius_inner ** 4 *(tail_length * radius_inner / radius_outer))
    I[1][1] = rho * (1 / 3 * np.pi * radius_outer ** 2 * tail_length ** 3 + np.pi / 20 * radius_outer ** 4 *tail_length - 1 / 3 * np.pi * radius_inner ** 2 * (tail_length * radius_inner / radius_outer) ** 3 + np.pi / 20 * radius_inner ** 4 *(tail_length * radius_inner / radius_outer))
    I[2][2] = rho * (np.pi /10 *radius_outer **4 *tail_length -np.pi /10 *radius_inner **4 *(tail_length * radius_inner / radius_outer)) # Izz

    # transform moment of inertia to global system   
    s        = np.array(center_of_gravity) - np.array(origin_cone) # vector from the cone base to the center of gravity of the aircraft. 
    I_global = np.array(I) + mass_cone * (np.array(np.dot(s[0], s[0])) * np.array(np.identity(3)) - s*np.transpose(s))
    
    # Add cone to the fuselage inertia tensor
    I_total = np.array(I_total) + np.array(I_global)
    
    return I_total,  mass_fuselage

def Volume_Fraction(radius_outer, radius_inner, center_length, tail_length):
    '''
    Calculate the volume fraction of each of the three components that make up the entire fuselage
    
    Assumptions:

    Source:

    Inputs:
    - Fuselage dimensions (radii, center length, tail length)

    Outputs:
    - Volume fraction of each fuselage component

    Properties Used:
    N/A
    '''
    # ----------------------------------------------------------------------------------------------------------------------    
    # Individual Component Volumes
    # ----------------------------------------------------------------------------------------------------------------------    
    
    volume_cone       = np.pi * radius_outer ** 2 * tail_length / 3 - np.pi * radius_inner ** 2 * (tail_length * radius_inner / radius_outer) / 3
    volume_hemisphere = 2 / 3 * np.pi * radius_outer ** 3 -2 / 3 * np.pi * radius_inner ** 3 
    volume_cylinder   = np.pi * center_length * (radius_outer ** 2 - radius_inner ** 2)
   
    # ----------------------------------------------------------------------------------------------------------------------    
    # Total Volume
    # ----------------------------------------------------------------------------------------------------------------------        
    volume_total = volume_cone + volume_hemisphere + volume_cylinder
    
    # ----------------------------------------------------------------------------------------------------------------------    
    # Volume Fractions
    # ----------------------------------------------------------------------------------------------------------------------        
    volume_fraction = np.array([volume_hemisphere / volume_total, volume_cylinder / volume_total, volume_cone / volume_total])
    
    return volume_fraction


# RCAIDE/Library/Methods/Propulsors/Converters/Ram/compute_ram_performance.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke    

# ----------------------------------------------------------------------------------------------------------------------
# compute_ram_performance
# ----------------------------------------------------------------------------------------------------------------------     
def compute_ram_performance(ram,ram_conditions,conditions):
    """ This computes the output values from the input values according T0
    equations from the source. The following properties are determined 
        conditions.freestream.
          stagnation_temperature              (numpy.ndarray): freestream stagnation_temperature             [K]
          stagnation_pressure                 (numpy.ndarray): freestream stagnation_pressure                [Pa]
          isentropic_expansion_factor         (numpy.ndarray): freestream isentropic_expansion_factor        [unitless]
          specific_heat_at_constant_pressure  (numpy.ndarray): freestream specific_heat_at_constant_pressure [J/(kg K)]
          gas_specific_constant               (numpy.ndarray): freestream gas_specific_constant              [J/(kg K)]
          speed_of_sound                      (numpy.ndarray): freestream speed_of_sound                     [m/s]
        ram.outputs.
          stagnation_temperature              (numpy.ndarray): exiting stagnation temperature     [K]
          stagnation_pressure                 (numpy.ndarray): exiting stagnation pressure        [Pa]
          isentropic_expansion_factor         (numpy.ndarray): isentropic expansion factor        [unitless]
          specific_heat_at_constant_pressure  (numpy.ndarray): specific_heat_at_constant_pressure [J/(kg K)]
          gas_specific_constant               (numpy.ndarray): gas_specific_constant              [J/(kg K)]   

    Assumptions:
        None

    Source:
        https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/

    Args:
        conditions.freestream.
           pressure               (numpy.ndarray): freestream pressure      [Pa]
           temperature            (numpy.ndarray): freestream temperature   [K]
           mach_number            (numpy.ndarray): freestream Mach number   [unitless]
        ram.inputs.working_fluid           (dict): working fluid properties [-]

    Returns:
        None 
  
    """
    # Unpack flight conditions 
    M0 = conditions.freestream.mach_number
    P0 = conditions.freestream.pressure
    T0 = conditions.freestream.temperature

    # Unpack ram inputs
    working_fluid  = ram.working_fluid
 
    # Compute the working fluid properties
    R        = working_fluid.gas_specific_constant
    gamma    = working_fluid.compute_gamma(T0,P0) 
    Cp       = working_fluid.compute_cp(T0,P0)
    a        = working_fluid.compute_speed_of_sound(T0,P0)
    V0       = a*M0 

    # Compute the stagnation quantities from the input static quantities
    stagnation_pressure    = P0*((1.+(gamma-1.)/2.*M0*M0 )**(gamma/(gamma-1.))) 
    stagnation_temperature = T0*(1.+((gamma-1.)/2.*M0*M0))

    # Store values into flight conditions data structure  
    conditions.freestream.isentropic_expansion_factor          = gamma
    conditions.freestream.specific_heat_at_constant_pressure   = Cp
    conditions.freestream.gas_specific_constant                = R
    conditions.freestream.stagnation_temperature               = stagnation_temperature
    conditions.freestream.stagnation_pressure                  = stagnation_pressure

    # Store values into compoment outputs  
    ram_conditions.outputs.isentropic_expansion_factor         = gamma
    ram_conditions.outputs.specific_heat_at_constant_pressure  = Cp
    ram_conditions.outputs.gas_specific_constant               = R
    ram_conditions.outputs.stagnation_temperature              = stagnation_temperature
    ram_conditions.outputs.stagnation_pressure                 = stagnation_pressure 
    ram_conditions.outputs.static_temperature                  = T0
    ram_conditions.outputs.static_pressure                     = P0
    ram_conditions.outputs.mach_number                         = M0
    ram_conditions.outputs.velocity                            = V0
    ram_conditions.outputs.speed_of_sound                      = a    
    
    return 


# RCAIDE/Library/Components/Propulsors/Converters/DC_Motor.py
# 
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
# RCAIDE imports  
from RCAIDE.Library.Components import Component
from RCAIDE.Library.Methods.Propulsors.Converters.DC_Motor.append_motor_conditions import  append_motor_conditions

# ----------------------------------------------------------------------------------------------------------------------
#  DC_Motor  
# ----------------------------------------------------------------------------------------------------------------------           
class DC_Motor(Component):
    """
    A direct current (DC) electric motor component model for electric propulsion systems.

    Attributes
    ----------
    tag : str
        Identifier for the motor. Default is 'motor'.
        
    resistance : float
        Internal electrical resistance of the motor [Ω]. Default is 0.0.
        
    no_load_current : float
        Current drawn by the motor with no mechanical load [A]. Default is 0.0.
        
    speed_constant : float
        Motor speed constant (Kv). Default is 0.0.
        
    rotor_radius : float
        Radius of the motor's rotor [m]. Default is 0.0.
        
    rotor_Cp : float
        Specific heat capacity of the rotor [J/kg/K]. Default is 0.0.
        
    efficiency : float
        Overall motor efficiency. Default is 1.0.
        
    gear_ratio : float
        Ratio of output shaft speed to motor speed. Default is 1.0.
        
    gearbox_efficiency : float
        Efficiency of the gearbox. Default is 1.0.
        
    expected_current : float
        Expected operating current [A]. Default is 0.0.
        
    power_split_ratio : float
        Ratio of power distribution when motor drives multiple loads. Default is 0.0.
        
    design_torque : float
        Design point torque output [N·m]. Default is 0.0.
        
    interpolated_func : callable
        Function for interpolating motor performance. Default is None.

    Notes
    -----
    The DC_Motor class models a direct current electric motor's performance
    characteristics. It accounts for electrical, mechanical, and thermal effects
    including:
    * Internal resistance losses
    * No-load current losses
    * Gearbox losses
    * Speed-torque relationships
    * Power distribution for multiple loads

    **Definitions**

    'Kv'
        Motor velocity constant, relating voltage to unloaded motor speed

    'No-load Current'
        Current drawn by motor to overcome internal friction when unloaded
        
    'Power Split Ratio'
        Fraction of total power delivered to primary load in multi-load applications

    See Also
    --------
    RCAIDE.Library.Methods.Propulsors.Converters.DC_Motor
    """      
    def __defaults__(self):
        """This sets the default values for the component to function.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        None
        """           
        self.tag                = 'motor' 
        self.resistance         = 0.0
        self.no_load_current    = 0.0
        self.speed_constant     = 0.0
        self.rotor_radius       = 0.0
        self.rotor_Cp           = 0.0
        self.efficiency         = 1.0
        self.gear_ratio         = 1.0
        self.gearbox_efficiency = 1.0
        self.expected_current   = 0.0
        self.power_split_ratio  = 0.0 
        self.design_torque      = 0.0 
        self.interpolated_func  = None
        
    def append_operating_conditions(self,segment,propulsor):
        propulsor_conditions =  segment.state.conditions.energy[propulsor.tag]
        append_motor_conditions(self,segment,propulsor_conditions)
        return
    


## @ingroup Methods-Propulsion
# fm_solver.py
#
# Created:  Sep 2017, P Goncalves
# Modified: Jan 2018, W. Maier

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

import numpy as np
from scipy.optimize import fsolve

# ----------------------------------------------------------------------
#  fm_solver
# ----------------------------------------------------------------------

## @ingroup Methods-Propulsion


def fm_solver(area_ratio, M0, gamma):
    """Function that takes in an area ratio and a Mach number associated to
    one of the areas and outputs the missing Mach number.

    Inputs:
    M           [-]
    gamma       [-]
    area_ratio  [-]

    Outputs:
    M1          [-]

    Source:
    https://web.stanford.edu/~cantwell/AA210A_Course_Material/AA210A_Course_Notes/
    """
    # Area-Mach Function
    func = lambda M1: ((M0/M1*((1.+(gamma-1.)/2.*M1*M1)/(1.+(gamma-1.)/2.*M0*M0))**((gamma+1.)/(2.*(gamma-1.))))-area_ratio)

    # Initializing the array
    M1_guess = np.ones_like(M0)

    # Separating supersonic and subsonic solutions
    i_low = M0 < 1.0
    i_high = M0 >= 1.0

    # Subsonic solution initialization
    M1_guess[i_low]= 0.1

    # Supersonic solution initialization
    M1_guess[i_high]= 1.1

    # Solving
    M1 = fsolve(func,M1_guess, factor=0.1)

    return M1



## @ingroup Core
# DataOrdered.py
#
# Created:  Jul 2016, E. Botero
# Modified: Sep 2016, E. Botero
#           May 2020, E. Botero
#           Jul 2020, E. Botero 
#           Jul 2021, E. Botero

   
# ----------------------------------------------------------------------
#   Imports
# ----------------------------------------------------------------------  

from collections import OrderedDict

# for enforcing attribute style access names
import string
chars = string.punctuation + string.whitespace
t_table = str.maketrans( chars          + string.ascii_uppercase , 
                            '_'*len(chars) + string.ascii_lowercase )

import numpy as np

# ----------------------------------------------------------------------
#   Property Class
# ----------------------------------------------------------------------   

class Property(object):
    """ Used to create the root map essential to the linking in DataOrdered()
       
        Assumptions:
        N/A
        
        Source:
        N/A
    """    
    
    def __init__(self,key=None):
        """ Initializes a property
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """           
        self._key = key
        
    def __get__(self,obj,kls=None):
        """ Gets a property
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            obj
    
            Outputs:
            self.key
    
            Properties Used:
            N/A    
        """           
        if obj is None: return self
        else          : return dict.__getitem__(obj,self._key)
        
    def __set__(self,obj,val):
        """ Sets a property
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            obj
            value
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """          
        dict.__setitem__(obj,self._key,val)
        
    def __delete__(self,obj):
        """ Deletes a property
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """          
        dict.__delitem__(obj,self._key)

    
# ----------------------------------------------------------------------
#   DataOrdered
# ----------------------------------------------------------------------        

## @ingroup Core
class DataOrdered(OrderedDict):
    """ An extension of the Python dict which allows for both tag and '.' usage.
        This is an ordered dictionary. So indexing it will produce deterministic results.
       
        Assumptions:
        N/A
        
        Source:
        N/A
    """
    
    
    _root = Property('_root')
    _map  = Property('_map')    
    
    def append(self,value,key=None):
        """ Adds new values to the classes. Can also change an already appended key
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            value
            key
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """         
        if key is None: key = value.tag
        key = key.translate(t_table)
        if key is None: key = value.tag
        if key in self: raise KeyError('key "%s" already exists' % key)
        self.__setattr__(key,value)    

    def __defaults__(self):
        """ A stub for all classes that come later
            
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """         
        pass
    
    def __getitem__(self,k):
        """ Retrieves an attribute set by a key k
            
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """          
        if not (isinstance(k,int) or isinstance(k,np.int64)):
            return super(DataOrdered,self).__getattribute__(k)
        else:
            return super(DataOrdered,self).__getattribute__(self.keys()[k])
    
    def __new__(cls,*args,**kwarg):
        """ Creates a new Data() class
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """         
        # Make the new:
        self = OrderedDict.__new__(cls)
        
        if self.hasattr('_root'):
            self._root
        else:
            root = [] # sentinel node
            root[:] = [root, root, None]
            dict.__setitem__(self,'_root',root)
            dict.__setitem__(self,'_map' ,{})        
        
        # Use the base init
        self.__init2()
        
        # get base class list
        klasses = self.get_bases()
                
        # fill in defaults trunk to leaf
        for klass in klasses[::-1]:
            klass.__defaults__(self)
            
        return self
    
    def hasattr(self,k):
        try:
            self.__getitem__(k)
            return True
        except:
            return False
            
    
    def __init__(self,*args,**kwarg):
        """ Initializes a new Data() class
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """         

        # handle input data (ala class factory)
        input_data = DataOrdered.__base__(*args,**kwarg)
        
        # update this data with inputs
        self.update(input_data)
        
        
    def __init2(self, items=None, **kwds):
        """ A helper that allows __init_ to complete the new Data() class
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """         
        def append_value(key,value):  
            
            self[key] = value            
        
        # a dictionary
        if hasattr(items, 'iterkeys'):
            for key in items.keys():
                append_value(key,items[key])

        elif hasattr(items, 'keys'):
            for key in items.keys():
                append_value(key,items[key])
                
        # items lists
        elif items:
            for key, value in items:
                append_value(key,value)
                
        # key words
        for key, value in kwds.items():
            append_value(key,value)     

    # iterate on values, not keys
    def __iter__(self):
        """ Returns all the iterable values. Can be used in a for loop.
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """          
        return iter(self.values())
    
    def get_bases(self):
        """ Finds the higher classes that may be built off of data
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            klasses
    
            Properties Used:
            N/A    
        """        
        # Get the Method Resolution Order, i.e. the ancestor tree
        klasses = list(self.__class__.__mro__)
        
        # Make sure that this is a Data object, otherwise throw an error.
        if DataOrdered not in klasses:
            raise TypeError('class %s is not of type Data()' % self.__class__)    
        
        # Remove the last two items, dict and object. Since the line before ensures this is a data object this won't break
        klasses = klasses[:-3]

        return klasses 
    
    def typestring(self):
        """ This function makes the .key.key structure in string form of Data()
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """           
        typestring = str(type(self)).split("'")[1]
        typestring = typestring.split('.')
        if typestring[-1] == typestring[-2]:
            del typestring[-1]
        typestring = '.'.join(typestring) 
        return typestring
    
    def dataname(self):
        """ This function is used for printing the class
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """        
        return "<data object '" + self.typestring() + "'>"

    def deep_set(self,keys,val):
        """ Regresses through a list of keys the same value in various places in a dictionary.
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            keys  - The keys to iterate over
            val   - The value to be set
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """         
        
        if isinstance(keys,str):
            keys = keys.split('.')
        
        data = self
         
        if len(keys) > 1:
            for k in keys[:-1]:
                data = data[k]
        
        data[ keys[-1] ] = val
        
        return data

    def deep_get(self,keys):
        """ Regresses through a list of keys to pull a specific value out
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            keys  - The keys to iterate over
            
            Outputs:
            value - The value to be retrieved
    
            Properties Used:
            N/A    
        """          
        
        if isinstance(keys,str):
            keys = keys.split('.')
        
        data = self
         
        if len(keys) > 1:
            for k in keys[:-1]:
                data = data[k]
        
        value = data[ keys[-1] ]
        
        return value   
    
    def update(self,other):
        """ Updates the internal values of a dictionary with given data
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            other
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """          
        if not isinstance(other,dict):
            raise TypeError('input is not a dictionary type')
        for k,v in other.items():
            # recurse only if self's value is a Dict()
            if k.startswith('_'):
                continue
        
            try:
                self[k].update(v)
            except:
                self[k] = v
        return 

    def __delattr__(self, key):
        """ An override of the standard __delattr_ in Python. This deletes whatever is called by k
            
            Assumptions:
            This one tries to treat k as an object, if that fails it treats it as a key.
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """            
        # Deleting an existing item uses self._map to find the link which is
        # then removed by updating the links in the predecessor and successor nodes.
        OrderedDict.__delattr__(self,key)
        link_prev, link_next, key = self._map.pop(key)
        link_prev[1] = link_next
        link_next[0] = link_prev 
        
    def __len__(self):
        """ This is overrides the Python function for checking length
            
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """          
        return self.__dict__.__len__()   

    def __iter_basic__(self):
        """ Returns all the iterable values. Can be used in a for loop.
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """           
        root = self._root
        curr = root[1]
        while curr is not root:
            yield curr[2]
            curr = curr[1]
            
    def __reduce__(self):
        """ Reduction function used for making configs
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """          
        items = [( k, DataOrdered.__getitem2(self,k) ) for k in DataOrdered.iterkeys(self)]
        inst_dict = vars(self).copy()
        for k in vars(DataOrdered()):
            inst_dict.pop(k, None)
        return (_reconstructor, (self.__class__,items,), inst_dict)
    
    def __setattr__(self, key, value):
        """ An override of the standard __setattr_ in Python.
            
            Assumptions:
            This one tries to treat k as an object, if that fails it treats it as a key.
    
            Source:
            N/A
    
            Inputs:
            key        [key]
            value        [value]
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """        
        # Setting a new item creates a new link which goes at the end of the linked
        # list, and the inherited dictionary is updated with the new key/value pair.
        if not hasattr(self,key) and not hasattr(self.__class__,key):
        #if not self.has_key(key) and not hasattr(self.__class__,key):
            root = dict.__getitem__(self,'_root')
            last = root[0]
            map  = dict.__getitem__(self,'_map')
            last[1] = root[0] = map[key] = [last, root, key]
        OrderedDict.__setattr__(self,key, value)

    def __setitem__(self,k,v):
        """ An override of the standard __setattr_ in Python.
            
            Assumptions:
            This one tries to treat k as an object, if that fails it treats it as a key.
    
            Source:
            N/A
    
            Inputs:
            k        [key]
            v        [value]
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """        
        self.__setattr__(k,v)
         

    def clear(self):
        """ Empties a dictionary
            
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """        
        
        try:
            for node in self._map.values():
                del node[:]
            root = self._root
            root[:] = [root, root, None]
            self._map.clear()
        except AttributeError:
            pass
        self.__dict__.clear()
        
    def get(self,k,d=None):
        """ Returns the values from k
            
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            k
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """         
        return self.__dict__.get(k,d)
        
    def has_key(self,k):
        """ Checks if the dictionary has the key, k
            
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            k
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """             
        return k in self.__dict__

    # allow override of iterators
    __iter = __iter__
    __getitem2 = OrderedDict.__getattribute__ 

    def keys(self):
        """ Returns a list of keys
            
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """         
        return list(self.__iter_basic__())
    
    def values(self):
        """ Returns all values inside the Data() class.
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            values
    
            Properties Used:
            N/A    
        """             
        return [self[key] for key in self.__iter_basic__()]
    
    def items(self):
        """ Returns all the items inside the data class
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            values
    
            Properties Used:
            N/A    
        """          
        return [(key, self[key]) for key in self.__iter_basic__()]
    
    def iterkeys(self):
        """ Returns all the keys which may be iterated over
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """         
        return self.__iter_basic__() 

# for rebuilding dictionaries with attributes
def _reconstructor(klass,items):
    """ For rebuilding dictionaries with attributes
        
        Assumptions:
        N/A

        Source:
        N/A

        Inputs:
        N/A

        Outputs:
        N/A

        Properties Used:
        N/A    
    """        
    self = DataOrdered.__new__(klass)
    DataOrdered.__init__(self,items)
    return self


# RCAIDE/Methods/Energy/Sources/battery_module/Lithium_Ion_NMC/.py
# 
# 
# Created:  Feb 2024, M. Clarke
# Modified: Sep 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Framework.Core                       import Units 
import numpy as np
from copy import  deepcopy
 
# ----------------------------------------------------------------------------------------------------------------------
# compute_nmc_cell_performance
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_nmc_cell_performance(battery_module,state,bus,coolant_lines,t_idx, delta_t): 
    """
    Compute the performance of a lithium-nickel-manganese-cobalt-oxide (NMC) battery_module cell.

    This function models the electrical and thermal behavior of an 18650 NMC battery_module cell
    based on experimental data from the Automotive Industrial Systems Company of Panasonic Group.

    Parameters
    ----------
    battery_module : battery_module
        The battery_module object containing cell properties and configuration.
    state : MissionState
        The current state of the mission.
    bus : ElectricBus
        The electric bus to which the battery_module is connected.
    coolant_lines : list
        List of coolant lines for thermal management.
    t_idx : int
        Current time index in the simulation.
    delta_t : float
        Time step size.

    Returns
    -------
    tuple
        A tuple containing:
        - stored_results_flag (bool): Flag indicating if results were stored.
        - stored_battery_module_tag (str): Tag of the battery_module for which results were stored.

    Notes
    -----
    The function updates various battery_module conditions in the `state` object, including:
    - Current energy
    - Temperature
    - Heat energy generated
    - Load power
    - Current
    - Open-circuit voltage
    - Charge throughput
    - Internal resistance
    - State of charge
    - Depth of discharge
    - Voltage under load

    The model includes:
    - Internal resistance calculation
    - Thermal modeling (heat generation and temperature change)
    - Electrical performance (voltage and current calculations)
    - State of charge and depth of discharge updates

    Arrays accessed from objects:
    - From bus_conditions:
        - power_draw
        - current_draw
    - From battery_module_conditions:
        - energy
        - voltage_open_circuit
        - cell.voltage_open_circuit
        - power
        - cell.power
        - internal_resistance
        - cell.internal_resistance
        - heat_energy_generated
        - cell.heat_energy_generated
        - voltage_under_load
        - cell.voltage_under_load
        - current
        - cell.current
        - temperature
        - cell.temperature
        - cell.state_of_charge
        - cell.energy
        - cell.charge_throughput
        - cell.depth_of_discharge

    References
    ----------
    .. [1] Zou, Y., Hu, X., Ma, H., and Li, S. E., "Combined State of Charge and State of
           Health estimation over lithium-ion battery_module cell cycle lifespan for electric 
           vehicles," Journal of Power Sources, Vol. 273, 2015, pp. 793-803.
           doi:10.1016/j.jpowsour.2014.09.146
    .. [2] Jeon, Dong Hyup, and Seung Man Baek. "Thermal modeling of cylindrical lithium ion 
           battery_module during discharge cycle." Energy Conversion and Management 52.8-9 (2011): 
           2973-2981.

    Assumptions
    -----------
    - All battery_module modules exhibit the same thermal behavior.
    - The cell temperature is assumed to be the temperature of the entire module.
    """

    # ---------------------------------------------------------------------------------    
    # battery cell properties
    # --------------------------------------------------------------------------------- 
    electrode_area            = battery_module.cell.electrode_area 
    As_cell                   = battery_module.cell.surface_area
    cell_mass                 = battery_module.cell.mass    
    Cp                        = battery_module.cell.specific_heat_capacity       
    battery_module_data       = battery_module.cell.discharge_performance_map
    
    # ---------------------------------------------------------------------------------
    # Compute Bus electrical properties 
    # ---------------------------------------------------------------------------------    
    bus_conditions              = state.conditions.energy[bus.tag]
    bus_config                  = bus.battery_module_electric_configuration
    E_bus                       = bus_conditions.energy
    P_bus                       = bus_conditions.power_draw
    I_bus                       = bus_conditions.current_draw
    
    # ---------------------------------------------------------------------------------
    # Compute battery_module Conditions
    # -------------------------------------------------------------------------    
    battery_module_conditions = state.conditions.energy[bus.tag].battery_modules[battery_module.tag]  
   
    E_module_max       = battery_module.maximum_energy * battery_module_conditions.cell.capacity_fade_factor
    
    V_oc_module        = battery_module_conditions.voltage_open_circuit
    V_oc_cell          = battery_module_conditions.cell.voltage_open_circuit   
  
    P_module           = battery_module_conditions.power
    P_cell             = battery_module_conditions.cell.power
    
    R_0_module         = battery_module_conditions.internal_resistance
    R_0_cell           = battery_module_conditions.cell.internal_resistance
    
    Q_heat_module      = battery_module_conditions.heat_energy_generated
    Q_heat_cell        = battery_module_conditions.cell.heat_energy_generated
    
    V_ul_module        = battery_module_conditions.voltage_under_load
    V_ul_cell          = battery_module_conditions.cell.voltage_under_load
    
    I_module           = battery_module_conditions.current 
    I_cell             = battery_module_conditions.cell.current
    
    T_module           = battery_module_conditions.temperature                 
    T_cell             = battery_module_conditions.cell.temperature
    
    SOC_cell           = battery_module_conditions.cell.state_of_charge  
    SOC_module         = battery_module_conditions.state_of_charge
    E_cell             = battery_module_conditions.cell.energy   
    E_module           = battery_module_conditions.energy
    Q_cell             = battery_module_conditions.cell.charge_throughput              
    DOD_cell           = battery_module_conditions.cell.depth_of_discharge
    
    # ---------------------------------------------------------------------------------
    # Compute battery_module electrical properties 
    # -------------------------------------------------------------------------    
    # Calculate the current going into one cell  
    n_series          = battery_module.electrical_configuration.series
    n_parallel        = battery_module.electrical_configuration.parallel 
    n_total           = n_series*n_parallel 
    no_modules        = bus.number_of_battery_modules
    
    # ---------------------------------------------------------------------------------
    # Examine Thermal Management System
    # ---------------------------------------------------------------------------------
    HAS = None  
    for coolant_line in coolant_lines:
        for tag, item in  coolant_line.items():
            if tag == 'battery_modules':
                for sub_tag, sub_item in item.items():
                    if sub_tag == battery_module.tag:
                        for btms in  sub_item:
                            HAS = btms    


    # ---------------------------------------------------------------------------------------------------
    # Current State 
    # ---------------------------------------------------------------------------------------------------
    if bus_config == 'Series':
        I_module[t_idx]      = I_bus[t_idx]
    elif bus_config  == 'Parallel':
        I_module[t_idx]      = I_bus[t_idx] / bus.number_of_battery_modules

    I_cell[t_idx] = I_module[t_idx] / n_parallel   
       
    # ---------------------------------------------------------------------------------
    # Compute battery_module cell temperature 
    # ---------------------------------------------------------------------------------
    R_0_cell[t_idx]                     =  (0.01483*(SOC_cell[t_idx]**2) - 0.02518*SOC_cell[t_idx] + 0.1036) *battery_module_conditions.cell.resistance_growth_factor  
    R_0_cell[t_idx][R_0_cell[t_idx]<0]  = 0. 

    # Determine temperature increase         
    sigma                 = 139 # Electrical conductivity
    n                     = 1
    F                     = 96485 # C/mol Faraday constant    
    delta_S               = -496.66*(SOC_cell[t_idx])**6 +  1729.4*(SOC_cell[t_idx])**5 + -2278 *(SOC_cell[t_idx])**4 +  1382.2 *(SOC_cell[t_idx])**3 + \
                            -380.47*(SOC_cell[t_idx])**2 +  46.508*(SOC_cell[t_idx])  + -10.692  

    i_cell                = I_cell[t_idx]/electrode_area # current intensity
    q_dot_entropy         = -(T_cell[t_idx])*delta_S*i_cell/(n*F)       
    q_dot_joule           = (i_cell**2)*(battery_module_conditions.cell.resistance_growth_factor)/(sigma)          
    Q_heat_cell[t_idx]    = (q_dot_joule + q_dot_entropy)*As_cell 
    Q_heat_module[t_idx]  = Q_heat_cell[t_idx]*n_total  

    V_ul_cell[t_idx]      = compute_nmc_cell_state(battery_module_data,SOC_cell[t_idx],T_cell[t_idx],abs(I_cell[t_idx])) 

    V_oc_cell[t_idx]      = V_ul_cell[t_idx] + (abs(I_cell[t_idx]) * R_0_cell[t_idx])              

    # Effective Power flowing through battery_module 
    P_module[t_idx]       = P_bus[t_idx] /no_modules  - np.abs(Q_heat_module[t_idx]) 

    # store remaining variables 
    V_oc_module[t_idx]     = V_oc_cell[t_idx]*n_series 
    V_ul_module[t_idx]     = V_ul_cell[t_idx]*n_series  
    T_module[t_idx]        = T_cell[t_idx]   # Assume the cell temperature is the temperature of the module
    P_cell[t_idx]          = P_module[t_idx]/n_total 
    E_module[t_idx]        = E_bus[t_idx]/no_modules 
    E_cell[t_idx]          = E_module[t_idx]/n_total  

    # ---------------------------------------------------------------------------------------------------     
    # Future State 
    # --------------------------------------------------------------------------------------------------- 
    if t_idx != state.numerics.number_of_control_points-1:  

        # Compute cell temperature
        if HAS is not None:
            T_cell[t_idx+1]  = HAS.compute_thermal_performance(battery_module,bus,coolant_line,Q_heat_cell[t_idx],T_cell[t_idx],state,delta_t[t_idx],t_idx)
        else:
            # Considers a thermally insulated system and the heat piles on in the system
            dT_dt              = Q_heat_cell[t_idx]/(cell_mass*Cp)
            T_cell[t_idx+1]    =  T_cell[t_idx] + dT_dt*delta_t[t_idx]
            
        # Compute state of charge and depth of discarge of the battery_module
        E_module[t_idx+1]                                     = (E_module[t_idx]) -P_module[t_idx]*delta_t[t_idx]
        E_module[t_idx+1][E_module[t_idx+1] > E_module_max]   = np.float32(E_module_max)
        SOC_cell[t_idx+1]                                     = E_module[t_idx+1]/E_module_max 
        SOC_cell[t_idx+1][SOC_cell[t_idx+1]>1]                = 1.
        SOC_cell[t_idx+1][SOC_cell[t_idx+1]<0]                = 0. 
        DOD_cell[t_idx+1]                                     = 1 - SOC_cell[t_idx+1]  
        SOC_module[t_idx+1]                                   = SOC_cell[t_idx+1]

    
        # Determine new charge throughput (the amount of charge gone through the battery_module)
        Q_cell[t_idx+1]    = Q_cell[t_idx] + abs(I_cell[t_idx])*delta_t[t_idx]/Units.hr
        
    stored_results_flag     = True
    stored_battery_module_tag     = battery_module.tag  
        
    return stored_results_flag, stored_battery_module_tag


def reuse_stored_nmc_cell_data(battery_module,state,bus,coolant_lines, t_idx, delta_t,stored_results_flag, stored_battery_module_tag):
    '''Reuses results from one propulsor for identical batteries
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    

    Outputs:  
    
    Properties Used: 
    N.A.        
    '''
   
    state.conditions.energy[bus.tag].battery_modules[battery_module.tag] = deepcopy(state.conditions.energy[bus.tag].battery_modules[stored_battery_module_tag])
    
        
    return
 
def compute_nmc_cell_state(battery_module_data,SOC,T,I):
    """This computes the electrical state variables of a lithium ion 
    battery_module cell with a  lithium-nickel-cobalt-aluminum oxide cathode 
    chemistry from look-up tables 
     
    Assumtions: 
    N/A
    
    Source:  
    N/A 
     
    Inputs:
        SOC           - state of charge of cell     [unitless]
        battery_module_data  - look-up data structure      [unitless]
        T             - battery_module cell temperature    [Kelvin]
        I             - battery_module cell current        [Amperes]
    
    Outputs:  
        V_ul          - under-load voltage          [Volts] 
        
    """ 

    # Make sure things do not break by limiting current, temperature and current 
    SOC[SOC < 0.]   = 0.  
    SOC[SOC > 1.]   = 1.    
    DOD             = 1 - SOC 
    
    T[np.isnan(T)] = 302.65
    T[T<272.65]    = 272.65 # model does not fit for below 0  degrees
    T[T>322.65]    = 322.65 # model does not fit for above 50 degrees
     
    I[I<0.0]       = 0.0
    I[I>8.0]       = 8.0   
     
    pts            = np.hstack((np.hstack((I, T)),DOD  )) # amps, temp, SOC   
    V_ul           = np.atleast_2d(battery_module_data.Voltage(pts)[:,1]).T  
    
    return V_ul


# RCAIDE/Library/Missions/Segments/Cruise/Constant_Dynamic_Pressure_Constant_Altitude_Loiter.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE 
import RCAIDE 

# Package imports  
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Constant dynamic pressure and constant altitude with set loiter time

    Source:
    N/A

    Inputs:
    segment.altitude                [meters]
    segment.time                    [seconds]
    segment.dynamic_pressure        [pascals]

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]
    conditions.frames.inertial.time             [seconds]

    Properties Used:
    N/A
    """     
    
    # unpack
    alt        = segment.altitude
    final_time = segment.time
    q          = segment.dynamic_pressure
    beta       = segment.sideslip_angle
    conditions = segment.state.conditions   
    
    # Update freestream to get density 
    RCAIDE.Library.Mission.Common.Update.atmosphere(segment) 
    rho        = conditions.freestream.density[:,0]   
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]       
    
    # check for initial velocity
    if q is None: 
        if not segment.state.initials: raise AttributeError('dynamic pressure not set')
        air_speed = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])
        
    else: # compute speed, constant with constant altitude
        air_speed = np.sqrt(q/(rho*0.5))
    
    # dimensionalize time
    t_initial = conditions.frames.inertial.time[0,0]
    t_final   = final_time + t_initial
    t_nondim  = segment.state.numerics.dimensionless.control_points
    time      = t_nondim * (t_final-t_initial) + t_initial
    v_x       = np.cos(beta)*air_speed 
    v_y       = np.sin(beta)*air_speed 
    
    # pack
    segment.state.conditions.freestream.altitude[:,0]             = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = v_x
    segment.state.conditions.frames.inertial.velocity_vector[:,1] = v_y
    segment.state.conditions.frames.inertial.time[:,0]            = time[:,0]


# RCAIDE/Library/Methods/Aerodynamics/Vortex_Lattice_Method/train_AVL_surrogates.py
#  
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
import RCAIDE 
from RCAIDE.Framework.Mission.Common                                             import Results  
from RCAIDE.Library.Methods.Aerodynamics.Athena_Vortex_Lattice.run_AVL_analysis  import run_AVL_analysis  
 
# Package imports 
import os
import numpy as np
from shutil import rmtree    

# ----------------------------------------------------------------------------------------------------------------------
#  train_AVL_surrogates
# ---------------------------------------------------------------------------------------------------------------------- 
def train_AVL_surrogates(aerodynamics):
    """Call methods to run VLM for sample point evaluation. 
    
    Assumptions:
        None
        
    Source:
        None

    Args:
        aerodynamics       : VLM analysis          [unitless] 
        
    Returns: 
        None    
    """ 
 
    run_folder             = os.path.abspath(aerodynamics.settings.filenames.run_folder)
    vehicle                = aerodynamics.vehicle
    training               = aerodynamics.training  
    AoA                    = training.angle_of_attack
    Mach                   = training.Mach
    side_slip_angle        = aerodynamics.settings.side_slip_angle
    roll_rate_coefficient  = aerodynamics.settings.roll_rate_coefficient
    pitch_rate_coefficient = aerodynamics.settings.pitch_rate_coefficient
    lift_coefficient       = aerodynamics.settings.lift_coefficient
    atmosphere             = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmo_data              = atmosphere.compute_values(altitude = 0.0)         
    
    len_AoA  = len(AoA)
    len_Mach = len(Mach)
    CM       = np.zeros((len_AoA,len_Mach))
    CL       = np.zeros_like(CM)
    CD       = np.zeros_like(CM)
    e        = np.zeros_like(CM)
    Cm_alpha = np.zeros_like(CM)
    Cn_beta  = np.zeros_like(CM)
    NP       = np.zeros_like(CM)  

    # remove old files in run directory  
    if os.path.exists(aerodynamics.settings.filenames.run_folder):
        if aerodynamics.settings.new_regression_results:
            rmtree(run_folder)

    for i,_ in enumerate(Mach):
        # Set training conditions
        run_conditions = Results()
        run_conditions.expand_rows(len_AoA)
        run_conditions.aerodynamics.angles.alpha           = np.array([AoA]).T  
        run_conditions.freestream.density                  = np.ones_like(run_conditions.aerodynamics.angles.alpha)*atmo_data.density 
        run_conditions.freestream.gravity                  = np.ones_like(run_conditions.aerodynamics.angles.alpha)*9.81          
        run_conditions.freestream.speed_of_sound           = np.ones_like(run_conditions.aerodynamics.angles.alpha)*atmo_data.speed_of_sound[0,0]  
        run_conditions.freestream.velocity                 = np.ones_like(run_conditions.aerodynamics.angles.alpha)*Mach[i] * run_conditions.freestream.speed_of_sound 
        run_conditions.freestream.mach_number              = np.ones_like(run_conditions.aerodynamics.angles.alpha)*Mach[i]
        run_conditions.aerodynamics.angles.beta            = np.ones_like(run_conditions.aerodynamics.angles.alpha)*side_slip_angle 
        run_conditions.static_stability.coefficients.roll  = np.ones_like(run_conditions.aerodynamics.angles.alpha)*roll_rate_coefficient   
        if lift_coefficient == None: 
            run_conditions.aerodynamics.coefficients.lift.total= lift_coefficient
        else:
            run_conditions.aerodynamics.coefficients.lift.total= np.array([lift_coefficient]).T  
        run_conditions.static_stability.coefficients.pitch = np.ones_like(run_conditions.aerodynamics.angles.alpha)*pitch_rate_coefficient 

        # Run Analysis at AoA[i] and Mach[i]
        run_AVL_analysis(aerodynamics,run_conditions)
 
        CL[:,i]       = run_conditions.aerodynamics.coefficients.lift.total[:,0]
        CD[:,i]       = run_conditions.aerodynamics.coefficients.drag.induced.total[:,0]      
        e [:,i]       = run_conditions.aerodynamics.coefficients.drag.induced.efficiency_factor[:,0]   
        CM[:,i]       = run_conditions.static_stability.coefficients.pitch[:,0]
        Cm_alpha[:,i] = run_conditions.static_stability.derivatives.CM_alpha[:,0]
        Cn_beta[:,i]  = run_conditions.static_stability.derivatives.CN_beta[:,0]
        NP[:,i]       = run_conditions.static_stability.neutral_point[:,0]     

    if aerodynamics.training_file:
        # load data 
        data_array   = np.loadtxt(aerodynamics.training_file) 
        
        # convert from 1D to 2D        
        CL_1D         = np.atleast_2d(data_array[:,0]) 
        CD_1D         = np.atleast_2d(data_array[:,1])            
        e_1D          = np.atleast_2d(data_array[:,2])
        CM_1D         = np.atleast_2d(data_array[:,3]) 
        Cm_alpha_1D   = np.atleast_2d(data_array[:,4])            
        Cn_beta_1D    = np.atleast_2d(data_array[:,5])
        NP_1D         = np.atleast_2d(data_array[:,6])

        # convert from 1D to 2D
        CL        = np.reshape(CL_1D, (len_AoA,-1))
        CD        = np.reshape(CD_1D, (len_AoA,-1))
        e         = np.reshape(e_1D , (len_AoA,-1)) 
        CM        = np.reshape(CM_1D, (len_AoA,-1))
        Cm_alpha  = np.reshape(Cm_alpha_1D, (len_AoA,-1))
        Cn_beta   = np.reshape(Cn_beta_1D , (len_AoA,-1))
        NP        = np.reshape(NP_1D , (len_AoA,-1))

    # Save the data for regression 
    if aerodynamics.settings.new_regression_results:
        # convert from 2D to 1D
        CL_1D       = CL.reshape([len_AoA*len_Mach,1]) 
        CD_1D       = CD.reshape([len_AoA*len_Mach,1])  
        e_1D        = e.reshape([len_AoA*len_Mach,1]) 
        CM_1D       = CM.reshape([len_AoA*len_Mach,1]) 
        Cm_alpha_1D = Cm_alpha.reshape([len_AoA*len_Mach,1])  
        Cn_beta_1D  = Cn_beta.reshape([len_AoA*len_Mach,1])         
        NP_1D       = Cn_beta.reshape([len_AoA*len_Mach,1]) 
        np.savetxt(vehicle.tag+'_stability_data.txt',np.hstack([CL_1D,CD_1D,e_1D,CM_1D,Cm_alpha_1D, Cn_beta_1D,NP_1D ]),fmt='%10.8f',header='   CM       Cm_alpha       Cn_beta       NP ')

    # Store training data
    # Save the data for regression
    training_data = np.zeros((7,len_AoA,len_Mach))
    training_data[0,:,:] = CL 
    training_data[1,:,:] = CD 
    training_data[2,:,:] = e  
    training_data[3,:,:] = CM       
    training_data[4,:,:] = Cm_alpha 
    training_data[5,:,:] = Cn_beta  
    training_data[6,:,:] = NP      

    # Store training data
    training.coefficients = training_data
    



# RCAIDE/Library/Methods/Stability/Dynamic_Stability/compute_dynamic_flight_modes.py
# 
# 
# Created:  Apr 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
import RCAIDE 
from RCAIDE.Library.Components.Wings.Control_Surfaces import Aileron , Elevator  

# python imports 
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  compute_dynamic_flight_modes
# ----------------------------------------------------------------------------------------------------------------------
def compute_dynamic_flight_modes(state,settings,aircraft): 
    """This function follows the stability axis EOM derivation in Blakelock
    to return the aircraft's dynamic modes and state space 
    
    Assumptions:
       Linerarized Equations are used following the reference below

    Source:
      Automatic Control of Aircraft and Missiles by J. Blakelock Pg 23 and 117
      Dynanmics of Flight Stability and Control by Bernard Edkin and Llyod Reid Chapter 5, 

    Inputs:
       conditions.aerodynamics  
       conditions.static_stability  
       conditions.stability.dynamic 

    Outputs: 
       conditions.dynamic_stability.LatModes
       conditions.dynamic_stability.LongModes 

    Properties Used:
       N/A
     """

    conditions = state.conditions 
    AoA        = conditions.aerodynamics.angles.alpha
    
    vertical_fligth_flag = False
    if isinstance(state,RCAIDE.Framework.Mission.Segments.Vertical_Flight.Climb) or \
       isinstance(state,RCAIDE.Framework.Mission.Segments.Vertical_Flight.Hover) or \
       isinstance(state,RCAIDE.Framework.Mission.Segments.Vertical_Flight.Descent):
        vertical_fligth_flag = True
    
    if (np.count_nonzero(aircraft.mass_properties.moments_of_inertia.tensor) > 0) and  (vertical_fligth_flag !=  True) and (np.all(np.isnan(AoA)) !=  True):
        g          = conditions.freestream.gravity  
        rho        = conditions.freestream.density
        u0         = conditions.freestream.velocity
        qDyn0      = conditions.freestream.dynamic_pressure  
        theta0     = np.arctan(conditions.frames.inertial.velocity_vector[:,2]/conditions.frames.inertial.velocity_vector[:,0])[:,None] 
        SS         = conditions.static_stability
        SSD        = SS.derivatives 
        DS         = conditions.dynamic_stability
                 
        num_cases  = len(AoA)
         
        b_ref  = conditions.b_ref
        c_ref  = conditions.c_ref
        S_ref  = conditions.S_ref  
        moments_of_inertia = aircraft.mass_properties.moments_of_inertia.tensor
        Ixx    = moments_of_inertia[0][0]
        Iyy    = moments_of_inertia[1][1]
        Izz    = moments_of_inertia[2][2]    
        if aircraft.mass_properties.mass == 0:
            m  = aircraft.mass_properties.max_takeoff
        elif aircraft.mass_properties.max_takeoff == 0:
            m = aircraft.mass_properties.mass
        else:
            raise AttributeError("Specify Vehicle Mass") 
        
        if np.all(conditions.static_stability.spiral_criteria) == 0: 
            conditions.static_stability.spiral_criteria = SSD.CL_beta*SSD.CN_r / (SSD.CL_r*SSD.CN_beta) 
          
        ## Build longitudinal EOM A Matrix (stability axis)
        ALon = np.zeros((num_cases,4,4))
        BLon = np.zeros((num_cases,4,1)) 
        CLon = np.zeros((num_cases,4,4))
        

        # Elevator effectiveness
        ht_tag         =  None
        main_wing_tag  = None
        for wing in aircraft.wings:
            if isinstance(wing,RCAIDE.Library.Components.Wings.Horizontal_Tail):
                ht_tag  = wing.tag
            if isinstance(wing,RCAIDE.Library.Components.Wings.Main_Wing):
                main_wing_tag = wing.tag
                
        if main_wing_tag != None and  ht_tag !=None: 
            main_wing       = aircraft.wings[main_wing_tag]     
            horizontal_tail = aircraft.wings[ht_tag] 
            
            # unpack unit conversions 
            V_t_prime       =  u0
            a_t             = 2 * np.pi # dCL_t_dalphat 
            l_t             = (horizontal_tail.origin[0][0] +horizontal_tail.aerodynamic_center[0]) - aircraft.mass_properties.center_of_gravity[0][0] # disstance from CG to tail AC
            S_t             = horizontal_tail.areas.reference # tail area
            S               = main_wing.areas.reference # wing area
            l_bar_t         = (horizontal_tail.origin[0][0] +  horizontal_tail.aerodynamic_center[0]) -(main_wing.origin[0][0] +  main_wing.aerodynamic_center[0])  # distance from AC of main wing to tail AC
            V_H             = ( l_bar_t *S_t ) /(c_ref *S)  # tail volume
            dEpsilon_dalpha =  0.3
            
            SSD.CZ_alpha_dot =  a_t * dEpsilon_dalpha *  (l_t /u0) *  (S_t / S) 
            SSD.CM_alpha_dot =  -a_t * V_H * dEpsilon_dalpha*  (l_t /u0)        
                        
        for i in range(num_cases): 
            CLon[i,:,:] = np.eye(4)        
        Cw         = m * g / (qDyn0 * S_ref)  
        Xu         = rho * u0 * S_ref * Cw * np.sin(theta0) + 0.5 * rho * u0 * S_ref * SSD.CX_u  
        Xw         = 0.5 * rho * u0 * S_ref * SSD.CX_alpha     
        Zu         = -rho * u0 * S_ref * Cw * np.cos(theta0) + 0.5 * rho * u0 * S_ref * SSD.CZ_u 
        Zw         = 0.5 * rho * u0 * S_ref * SSD.CZ_alpha   
        Zq         = 0.25 * rho * u0 * c_ref * S_ref *  SSD.CZ_q    
        Mu         =  0.5 * rho * u0 * c_ref * S_ref * SSD.CM_u     
        Mw         = 0.5 * rho * u0 * c_ref * S_ref * SSD.CM_alpha  
        Mq         = 0.25 * rho * u0 * c_ref * c_ref * S_ref * SSD.CM_q     
        ZwDot      = 0.25 * rho * c_ref * S_ref * SSD.CZ_alpha_dot  
        MwDot      = 0.25 * rho * c_ref * S_ref * SSD.CM_alpha_dot  
        
        
        ALon[:,0,0] = (Xu / m).T[0]
        ALon[:,0,1] = (Xw / m).T[0] 
        ALon[:,0,3] = (-g * np.cos(theta0)).T[0]
        ALon[:,1,0] = (Zu / (m - ZwDot)).T[0]
        ALon[:,1,1] = (Zw / (m - ZwDot)).T[0] 
        ALon[:,1,2] = ((Zq + (m * u0)) / (m - ZwDot) ).T[0]
        ALon[:,2,0] = ((Mu + MwDot * Zu / (m - ZwDot)) / Iyy).T[0]  
        ALon[:,2,1] = ((Mw + MwDot * Zw / (m - ZwDot)) / Iyy).T[0] 
        ALon[:,2,2] = ((Mq + MwDot * (Zq + m * u0) / (m - ZwDot)) / Iyy ).T[0] 
        ALon[:,2,3] = (-MwDot * m * g * np.sin(theta0) / (Iyy * (m - ZwDot))).T[0] 
        ALon[:,3,0] = 0
        ALon[:,3,1] = 0
        ALon[:,3,2] = 1
        ALon[:,3,3] = 0 

        for wing in aircraft.wings: 
            if wing.control_surfaces :
                for ctrl_surf in wing.control_surfaces: 
                    if (type(ctrl_surf) ==  Elevator):
                        ele = conditions.control_surfaces.elevator.static_stability.coefficients 
                        Xe  = 0 # Neglect
                        Ze  = 0.5 * rho * u0 * u0 * S_ref * ele.lift
                        Me  = 0.5 * rho * u0 * u0 * S_ref * c_ref * ele.M
                        
                        BLon[:,0,0] = Xe / m
                        BLon[:,1,0] = (Ze / (m - ZwDot)).T[0]
                        BLon[:,2,0] = (Me / Iyy + MwDot / Iyy * Ze / (m - ZwDot)).T[0]
                        BLon[:,3,0] = 0        
         
        # Look at eigenvalues and eigenvectors
        LonModes                  = np.zeros((num_cases,4), dtype = complex)
        phugoidFreqHz             = np.zeros((num_cases,1))
        phugoidDamping            = np.zeros((num_cases,1))
        phugoidTimeDoubleHalf     = np.zeros((num_cases,1))
        shortPeriodFreqHz         = np.zeros((num_cases,1))
        shortPeriodDamping        = np.zeros((num_cases,1))
        shortPeriodTimeDoubleHalf = np.zeros((num_cases,1))
        
        if np.any(np.isnan(ALon)):
            pass
        else:
            for i_long in range(num_cases):
                D  , V = np.linalg.eig(ALon[i_long,:,:]) # State order: u, w, q, theta
                LonModes[i_long,:] = D
                
                # Find phugoid
                phugoidInd                    = np.argmax(D)  
                phugoidFreqHz[i_long]         = abs(D[phugoidInd]) / (2 * np.pi)
                phugoidDamping[i_long]        = np.sqrt(1/ (1 + ( D[phugoidInd].imag/ D[phugoidInd].real )**2 ))  
                phugoidTimeDoubleHalf[i_long] = np.log(2) / abs(2 * np.pi * phugoidFreqHz[i_long] * phugoidDamping[i_long])
                
                # Find short period
                shortPeriodInd                    = np.argmin(D)  
                shortPeriodFreqHz[i_long]         = abs(D[shortPeriodInd]) / (2 * np.pi)
                shortPeriodDamping[i_long]        = np.sqrt(1/ (1 + (D[shortPeriodInd].imag/D[shortPeriodInd].real)**2 ))  
                shortPeriodTimeDoubleHalf[i_long] = np.log(2) / abs(2 * np.pi * shortPeriodFreqHz[i_long] * shortPeriodDamping[i_long]) 
        
        ## Build lateral EOM A Matrix (stability axis)
        ALat = np.zeros((num_cases,4,4))
        BLat = np.zeros((num_cases,4,1))
        CLat = np.zeros((num_cases,4,4))
        for c_i1 in range(num_cases): 
            CLat[c_i1,:,:] = np.eye(4) 
        DLat = np.zeros((num_cases,4,1))
        
        # Need to compute Ixx, Izz, and Ixz as a function of alpha
        Ixp  = np.zeros((num_cases,1))
        Izp  = np.zeros((num_cases,1))
        Ixzp = np.zeros((num_cases,1))
        
        for c_i2 in range(num_cases): 
            R       = np.array( [[np.cos(AoA[c_i2][0]) ,  - np.sin(AoA[c_i2][0]) ], [ np.sin( AoA[c_i2][0]) , np.cos(AoA[i][0])]])
            modI    = np.array([[moments_of_inertia[0][0],moments_of_inertia[0][2]],[moments_of_inertia[2][0],moments_of_inertia[2][2]]] ) 
            INew    = R * modI  * np.transpose(R)
            IxxStab =  INew[0,0]
            IxzStab = -INew[0,1]
            IzzStab =  INew[1,1]
            Ixp[c_i2]  = (IxxStab * IzzStab - IxzStab**2) / IzzStab
            Izp[c_i2]  = (IxxStab * IzzStab - IxzStab**2) / IxxStab
            Ixzp[c_i2] = IxzStab / (IxxStab * IzzStab - IxzStab**2) 
            
        Yv = 0.5 * rho * u0 * S_ref * SSD.CY_beta 
        Yr = 0.25 * rho * u0 * b_ref * S_ref * SSD.CY_r
        Lv = 0.5 * rho * u0 * b_ref * S_ref * SSD.CL_beta
        Lp = 0.25 * rho * u0 * b_ref**2 * S_ref * SSD.CL_p
        Lr = 0.25 * rho * u0 * b_ref**2 * S_ref * SSD.CL_r
        Nv = 0.5 * rho * u0 * b_ref * S_ref * SSD.CN_beta
        Np = 0.25 * rho * u0 * b_ref**2 * S_ref * SSD.CN_p
        Nr = 0.25 * rho * u0 * b_ref**2 * S_ref * SSD.CN_r
        
        # Aileron effectiveness 
        for wing in aircraft.wings:
            if wing.control_surfaces :
                for ctrl_surf in wing.control_surfaces:
                    if (type(ctrl_surf) ==  Aileron): 
                        ail = conditions.control_surfaces.aileron.static_stability.coefficients                  
                        Ya = 0.5 * rho * u0 * u0 * S_ref * ail.Y 
                        La = 0.5 * rho * u0 * u0 * S_ref * b_ref * ail.L 
                        Na = 0.5 * rho * u0 * u0 * S_ref * b_ref * ail.N 
                        
                        BLat[:,0,0] = (Ya / m).T[0]
                        BLat[:,1,0] = (La / Ixp + Ixzp * Na).T[0]
                        BLat[:,2,0] = (Ixzp * La + Na / Izp).T[0]
                        BLat[:,3,0] = 0
     
        ALat[:,0,0] = (Yv / m).T[0]  
        ALat[:,0,2] = (Yr/m - u0).T[0] 
        ALat[:,0,3] = (g * np.cos(theta0)).T[0]
        
        ALat[:,1,0] = (Lv / Ixp + Ixzp * Nv).T[0] 
        ALat[:,1,1] = (Lp / Ixp + Ixzp * Np).T[0] 
        ALat[:,1,2] = (Lr / Ixp + Ixzp * Nr).T[0] 
        ALat[:,1,3] = 0
        
        ALat[:,2,0] = (Ixzp * Lv + Nv / Izp).T[0] 
        ALat[:,2,1] = (Ixzp * Lp + Np / Izp).T[0] 
        ALat[:,2,2] = (Ixzp * Lr + Nr / Izp).T[0] 
        ALat[:,2,3] = 0
        
        ALat[:,3,0] = 0
        ALat[:,3,1] = 1
        ALat[:,3,2] = (np.tan(theta0)).T[0] 
        ALat[:,3,3] = 0
                                    
        LatModes                    = np.zeros((num_cases,4),dtype=complex)
        dutchRollFreqHz             = np.zeros((num_cases,1))
        dutchRollDamping            = np.zeros((num_cases,1))
        dutchRollTimeDoubleHalf     = np.zeros((num_cases,1))
        rollSubsistenceFreqHz       = np.zeros((num_cases,1))
        rollSubsistenceTimeConstant = np.zeros((num_cases,1))
        rollSubsistenceDamping      = np.zeros((num_cases,1))
        spiralFreqHz                = np.zeros((num_cases,1))
        spiralTimeDoubleHalf        = np.zeros((num_cases,1))
        spiralDamping               = np.zeros((num_cases,1))
        dutchRoll_mode_real         = np.zeros((num_cases,1))
        
        for i_lat in range(num_cases):
            try: 
                D  , V = np.linalg.eig(ALat[i_lat,:,:]) # State order: u, w, q, theta
                LatModes[i_lat,:] = D  
                
                real_parts = LatModes[i_lat,:].real
                unique_elements, counts = np.unique(real_parts, return_counts=True)
                idx = np.where(counts==2)[0]
    
                dutchRollFreqHz[i_lat]         = abs(D[idx]) / (2 * np.pi)
                dutchRollDamping[i_lat]        = np.sqrt(1/ (1 + ( D[idx].imag/ D[idx].real )**2 ))  
                dutchRollTimeDoubleHalf[i_lat] = np.log(2) / abs(2 * np.pi * dutchRollFreqHz[i_lat] * dutchRollDamping[i_lat])
                dutchRoll_mode_real[i_lat]     = D[idx].real / (2 * np.pi)
                
                dutch_roll_idx                 = np.where( unique_elements[idx] == D.real )[0]
                
                remaining_modes                    = np.delete(D, dutch_roll_idx)
                rollInd                            = np.argmin(remaining_modes) 
                rollSubsistenceFreqHz[i_lat]       = abs(remaining_modes[rollInd]) / 2 / np.pi
                rollSubsistenceDamping[i_lat]      = - np.sign(remaining_modes[rollInd].real)
                rollSubsistenceTimeConstant[i_lat] = 1 / (2 * np.pi * rollSubsistenceFreqHz[i_lat] * rollSubsistenceDamping[i_lat])
                
                # Find spiral mode 
                sprial_mode                 = np.delete(remaining_modes,rollInd)             
                spiralFreqHz[i_lat]         = abs(sprial_mode) / 2 / np.pi
                spiralDamping[i_lat]        = - np.sign(sprial_mode.real)
                spiralTimeDoubleHalf[i_lat] = np.log(2) / abs(2 * np.pi * spiralFreqHz[i_lat] * spiralDamping[i_lat])
            except:
                pass 
        
        # Inertial coupling susceptibility
        # See Etkin & Reid pg. 118 
        DS.pMax = min(min(np.sqrt(-Mw * u0 / (Izz - Ixx))), min(np.sqrt(-Nv * u0 / (Iyy - Ixx)))) 
        
        # -----------------------------------------------------------------------------------------------------------------------  
        # Store Results
        # ------------------------------------------------------------------------------------------------------------------------  
        DS.LongModes.LongModes                    = LonModes
        #DS.LongModes.LongSys                      = LonSys    
        DS.LongModes.phugoidFreqHz                = phugoidFreqHz
        DS.LongModes.phugoidDamping               = phugoidDamping
        DS.LongModes.phugoidTimeDoubleHalf        = phugoidTimeDoubleHalf
        DS.LongModes.shortPeriodFreqHz            = shortPeriodFreqHz
        DS.LongModes.shortPeriodDamping           = shortPeriodDamping
        DS.LongModes.shortPeriodTimeDoubleHalf    = shortPeriodTimeDoubleHalf
                                                                        
        DS.LatModes.LatModes                      = LatModes  
        #DS.LatModes.Latsys                        = LatSys   
        DS.LatModes.dutchRollFreqHz               = dutchRollFreqHz
        DS.LatModes.dutchRollDamping              = dutchRollDamping
        DS.LatModes.dutchRollTimeDoubleHalf       = dutchRollTimeDoubleHalf
        DS.LatModes.dutchRoll_mode_real           = dutchRoll_mode_real 
        DS.LatModes.rollSubsistenceFreqHz         = rollSubsistenceFreqHz
        DS.LatModes.rollSubsistenceTimeConstant   = rollSubsistenceTimeConstant
        DS.LatModes.rollSubsistenceDamping        = rollSubsistenceDamping
        DS.LatModes.spiralFreqHz                  = spiralFreqHz
        DS.LatModes.spiralTimeDoubleHalf          = spiralTimeDoubleHalf 
        DS.LatModes.spiralDamping                 = spiralDamping
    
    return 



# RCAIDE/Methods/Energy/Propulsors/Turbojet_Propulsor/compute_thrust.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports  
from RCAIDE.Framework.Core      import Units 

# Python package imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  compute_thrust
# ----------------------------------------------------------------------------------------------------------------------
def compute_thrust(turbojet,turbojet_conditions,conditions):
    """Computes thrust and other properties as below.

    Assumptions:
    Perfect gas

    Source:
    https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/

    Inputs:
    conditions.freestream.
      isentropic_expansion_factor        [-] (gamma)
      specific_heat_at_constant_pressure [J/(kg K)]
      velocity                           [m/s]
      speed_of_sound                     [m/s]
      mach_number                        [-]
      pressure                           [Pa]
      gravity                            [m/s^2]
    conditions.throttle                  [-] (.1 is 10%)
    turbojet_conditions.
      fuel_to_air_ratio                  [-]
      total_temperature_reference        [K]
      total_pressure_reference           [Pa]
      core_nozzle.
        velocity                         [m/s]
        static_pressure                  [Pa]
        area_ratio                       [-]
      fan_nozzle.
        velocity                         [m/s]
        static_pressure                  [Pa]
        area_ratio                       [-]
      number_of_engines                  [-]
      bypass_ratio                       [-]
      flow_through_core                  [-] percentage of total flow (.1 is 10%)
      flow_through_fan                   [-] percentage of total flow (.1 is 10%)

    Outputs:
    turbojet_conditions.
      thrust                             [N]
      thrust_specific_fuel_consumption   [N/N-s]
      non_dimensional_thrust             [-]
      core_mass_flow_rate                [kg/s]
      fuel_flow_rate                     [kg/s]
      power                              [W]
      Specific Impulse                   [s]

    Properties Used:
    turbojet.
      reference_temperature              [K]
      reference_pressure                 [Pa]
      compressor_nondimensional_massflow [-]
      SFC_adjustment                     [-]
    """           
    #unpack the values

    #unpacking from conditions
    gamma                       = conditions.freestream.isentropic_expansion_factor 
    u0                          = conditions.freestream.velocity
    a0                          = conditions.freestream.speed_of_sound
    M0                          = conditions.freestream.mach_number
    p0                          = conditions.freestream.pressure  
    g                           = conditions.freestream.gravity        

    #unpacking from inputs
    Tref                        = turbojet.reference_temperature
    Pref                        = turbojet.reference_pressure
    mdhc                        = turbojet.compressor_nondimensional_massflow
    SFC_adjustment              = turbojet.SFC_adjustment 
    f                           = turbojet_conditions.fuel_to_air_ratio
    total_temperature_reference = turbojet_conditions.total_temperature_reference
    total_pressure_reference    = turbojet_conditions.total_pressure_reference   
    core_area_ratio             = turbojet_conditions.core_nozzle_area_ratio  
    V_core_nozzle               = turbojet_conditions.core_nozzle_exit_velocity
    P_core_nozzle               = turbojet_conditions.core_nozzle_static_pressure     
    flow_through_core           = turbojet_conditions.flow_through_core  
 
    #computing the non dimensional thrust
    core_thrust_nondimensional  = flow_through_core*(gamma*M0*M0*(V_core_nozzle/u0-1.) + core_area_ratio*( P_core_nozzle/p0-1.)) 

    Thrust_nd                   = core_thrust_nondimensional  

    #Computing Specifc Thrust
    Fsp              = 1./(gamma*M0)*Thrust_nd

    #Computing the specific impulse
    Isp              = Fsp*a0/(f*g)

    #Computing the TSFC
    TSFC             = f*g/(Fsp*a0)*(1.-SFC_adjustment) * Units.hour # 1/s is converted to 1/hr here

    #computing the core mass flow
    mdot_core        = mdhc*np.sqrt(Tref/total_temperature_reference)*(total_pressure_reference/Pref)

    #computing the dimensional thrust
    FD2              = Fsp*a0*mdot_core* turbojet_conditions.throttle

    #fuel flow rate
    a = np.array([0.])        
    fuel_flow_rate   = np.fmax(FD2*TSFC/g,a)*1./Units.hour

    #computing the power 
    power            = FD2*u0

    # pack outputs 
    turbojet_conditions.thrust                            = FD2 
    turbojet_conditions.thrust_specific_fuel_consumption  = TSFC
    turbojet_conditions.non_dimensional_thrust            = Fsp 
    turbojet_conditions.core_mass_flow_rate               = mdot_core
    turbojet_conditions.fuel_flow_rate                    = fuel_flow_rate    
    turbojet_conditions.power                             = power  
    turbojet_conditions.specific_impulse                  = Isp

    return 


#  RCAIDE/Library/Methods/Emissions/Chemical_Reactor_Network_Method/train_CRN_EI_surrogates.py
#  
#  Created: 2024, M. Clarke, M. Guidotti

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
import RCAIDE 
from RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.evaluate_cantera import evaluate_cantera 

# package imports    
import numpy    as np  

# ----------------------------------------------------------------------------------------------------------------------
#  Train Cantera Model 
# ----------------------------------------------------------------------------------------------------------------------
def train_CRN_EI_surrogates(emissions): 
    """
    Generates training data for emission index surrogate models using Chemical Reactor Network simulations.

    Parameters
    ----------
    emissions : Data
        Container for emissions data and settings
        
        - training : Data
            Training data container

            - pressure : ndarray
                Array of pressure values to evaluate [Pa]
            - temperature : ndarray
                Array of temperature values to evaluate [K]
            - air_mass_flowrate : ndarray
                Array of air mass flow rates to evaluate [kg/s]
            - fuel_to_air_ratio : ndarray
                Array of fuel-to-air ratios to evaluate [-]
        
        - vehicle : Data
            Vehicle configuration data

            - networks : list
                List of propulsion system networks

                - propulsors : list
                    Propulsion units containing combustors

    Returns
    -------
    Updates emissions.training with:

    EI_CO2 : ndarray
        CO2 emission index training data [kg_CO2/kg_fuel]
    EI_CO : ndarray
        CO emission index training data [kg_CO/kg_fuel]
    EI_H2O : ndarray
        H2O emission index training data [kg_H2O/kg_fuel]
    EI_NO : ndarray
        NO emission index training data [kg_NO/kg_fuel]
    EI_NO2 : ndarray
        NO2 emission index training data [kg_NO2/kg_fuel]

    Notes
    -----
    This function evaluates emission indices across a grid of operating conditions
    to generate training data for surrogate models. The grid is formed by the
    combinations of the input arrays (pressure, temperature, mass flow, and
    fuel-to-air ratio).

    **Extra modules required:**

    * numpy
    * Cantera 

    **Major Assumptions**

    * Operating points form a regular grid
    * All combinations of input parameters are valid

    **Theory**
    Training data is generated by:

    1. Creating a grid of operating points
    2. Evaluating detailed chemical kinetics at each point
    3. Storing resulting emission indices in 4D arrays
    
    The grid dimensions are:
    :math:`[P] × [T] × [\dot{m}] × [FAR]`

    Where:

    * :math:`P` = Pressure points
    * :math:`T` = Temperature points
    * :math:`\dot{m}` = Mass flow rate points
    * :math:`FAR` = Fuel-to-air ratio points

    See Also
    --------
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.evaluate_cantera 
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.build_CRN_EI_surrogates
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.evaluate_CRN_emission_indices

    References
    ----------
    [1] Goodwin, D. G., Speth, R. L., Moffat, H. K., & Weber, B. W. (2023). Cantera: An object-oriented software toolkit for chemical kinetics,thermodynamics, and transport processes (Version 3.0.0) [Computer software]. https://www.cantera.org
    """
    # unpack data 
    P              = emissions.training.pressure         
    T              = emissions.training.temperature      
    mdot           = emissions.training.air_mass_flowrate
    FAR            = emissions.training.fuel_to_air_ratio
    
    vehicle        = emissions.vehicle
    for network in vehicle.networks:   
        for propulsor in  network.propulsors:
            if  isinstance(propulsor,RCAIDE.Library.Components.Propulsors.Turbofan) or \
                isinstance(propulsor,RCAIDE.Library.Components.Propulsors.Turbojet) or \
                isinstance(propulsor,RCAIDE.Library.Components.Propulsors.Turboshaft) or \
                isinstance(propulsor,RCAIDE.Library.Components.Propulsors.ICE_Propeller):            
                combustor = propulsor.combustor              
            else:
                combustor = False 
            
    len_P    = len(P)
    len_T    = len(T)
    len_mdot = len(mdot)
    len_far  = len(FAR) 
    
    EI_CO2 = np.zeros((len_P,len_T,len_mdot,len_far))
    EI_CO  = np.zeros((len_P,len_T,len_mdot,len_far))
    EI_H2O = np.zeros((len_P,len_T,len_mdot,len_far))
    EI_NO2 = np.zeros((len_P,len_T,len_mdot,len_far))
    EI_NO  = np.zeros((len_P,len_T,len_mdot,len_far))  
    
    if combustor == False:
        emissions.no_combustor = True
        return 
    
    for p_i in range(len_P):
        for t_i in range(len_T):
            for mdot_i in  range(len_mdot):
                for far_i in  range(len_far):
                    
                    # Call cantera 
                    results = evaluate_cantera(combustor,T[t_i],P[p_i],mdot[mdot_i],FAR[far_i]) 
                    
                    EI_CO2[p_i, t_i, mdot_i,far_i] = results.EI_CO2
                    EI_CO [p_i, t_i, mdot_i,far_i] = results.EI_CO 
                    EI_H2O[p_i, t_i, mdot_i,far_i] = results.EI_H2O
                    EI_NO [p_i, t_i, mdot_i,far_i] = results.EI_NO 
                    EI_NO2[p_i, t_i, mdot_i,far_i] = results.EI_NO2
    
    emissions.training.EI_CO2 = EI_CO2
    emissions.training.EI_CO =  EI_CO
    emissions.training.EI_H2O = EI_H2O
    emissions.training.EI_NO =  EI_NO
    emissions.training.EI_NO2 = EI_NO2
    
    return 


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/General_Aviation/compute_systems_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
from RCAIDE.Framework.Core import  Units , Data 

# ----------------------------------------------------------------------------------------------------------------------
# Systems Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_systems_weight(W_uav, V_fuel, V_int, N_tank, N_eng, l_fuselage, span, TOW, Nult, num_seats,  mach_number, has_air_conditioner=1):
    """ output = RCAIDE.Methods.Weights.Correlations.General_Avation.systems(num_seats, ctrl_type, S_h, S_v, S_gross_w, ac_type)
        Calculate the weight of the different engine systems on the aircraft

        Source:
            Raymer, Aircraft Design: A Conceptual Approach (pg 461 in 4th edition)

        Inputs:
            V_fuel              - total fuel volume                     [meters**3]
            V_int               - internal fuel volume                  [meters**3]
            N_tank              - number of fuel tanks                  [dimensionless]
            N_eng               - number of engines                     [dimensionless]
            span                - wingspan                              [meters]
            TOW                 - gross takeoff weight of the aircraft  [kg]
            num_seats           - total number of seats on the aircraft [dimensionless]
            mach_number         - mach number                           [dimensionless]
            has_air_conditioner - integer of 1 if the vehicle has ac, 0 if not

        Outputs:
            output - a data dictionary with fields:
                W_flight_controls - weight of the flight control system [kilograms]
                W_apu - weight of the apu [kilograms]
                W_hyd_pnu - weight of the hydraulics and pneumatics [kilograms]
                W_avionics - weight of the avionics [kilograms]
                W_opitems - weight of the optional items based on the type of aircraft [kilograms]
                W_electrical - weight of the electrical items [kilograms]
                W_ac - weight of the air conditioning and anti-ice system [kilograms]
                W_furnish - weight of the furnishings in the fuselage [kilograms]
    """ 
    # unpack inputs 
    Q_tot  = V_fuel/Units.gallons
    Q_int  = V_int/Units.gallons 
    l_fus  = l_fuselage / Units.ft  # Convert meters to ft
    b_wing = span/Units.ft 
    W_0    = TOW/Units.lb
    
    # Fuel system
    W_fuel_system = 2.49*(Q_tot**.726)*((Q_tot/(Q_tot+Q_int))**.363)*(N_tank**.242)*(N_eng**.157)*Units.lb

    # Flight controls
    W_flight_controls = .053*(l_fus**1.536)*(b_wing**.371)*((Nult*W_0**(10.**(-4.)))**.8)*Units.lb
    
    # Hydraulics & Pneumatics Group Wt
    hyd_pnu_wt = (.001*W_0) * Units.lb

    # Avionics weight
    W_avionics = 2.117*((W_uav/Units.lbs)**.933)*Units.lb 

    # Electrical Group Wt
    W_electrical = 12.57*((W_avionics/Units.lb + W_fuel_system/Units.lb)**.51)*Units.lb

    # Environmental Control 
    W_air_conditioning = has_air_conditioner*.265*(W_0**.52)*((1. * num_seats)**.68)*((W_avionics/Units.lb)**.17)*(mach_number**.08)*Units.lb

    # Furnishings Group Wt
    W_furnish = (.0582*W_0-65.)*Units.lb

    # packup outputs
    output = Data()   
    output.W_flight_control    = W_flight_controls
    output.W_hyd_pnu           = hyd_pnu_wt
    output.W_avionics          = W_avionics
    output.W_electrical        = W_electrical
    output.W_ac                = W_air_conditioning
    output.W_furnish           = W_furnish
    output.W_fuel_system       = W_fuel_system
    output.total               = output.W_flight_control + output.W_hyd_pnu \
                                  + output.W_ac + output.W_avionics + output.W_electrical \
                                  + output.W_furnish + output.W_fuel_system

    return output


# RCAIDE/Library/Missions/Common/Initialize/weights.py
# 
# 
# Created:  Jul 2023, M. Clarke
 

# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Weights
# ---------------------------------------------------------------------------------------------------------------------- 
def weights(segment):
    """ Initializes weight of vehicle
    
        Assumptions:   
            Uses max takeoff weight if no weight analysis is performed 
            
        Inputs: 
            state.
                initials.conditions.weights.total_mass  [kg]
            state.conditions:           
                conditions.weights.total_mass           [kg]
            
        Outputs: 
            state.conditions.weights.total_mass         [kg] 
           
        Properties Used:
        N/A 
    """      
 
    if segment.state.initials:
        m_initial = segment.state.initials.conditions.weights.total_mass[-1,0] 
    else: 
        if segment.analyses.weights != None: 
            m_initial = segment.analyses.weights.vehicle.mass_properties.takeoff
        else: 
            m_initial = segment.analyses.energy.vehicle.networks[list(segment.analyses.energy.vehicle.networks.keys())[0]].mass_properties.mass

    m_current = segment.state.conditions.weights.total_mass
    
    segment.state.conditions.weights.total_mass[:,:] = m_current + (m_initial - m_current[0,0])
        
    return 


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_vertical_tail_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core    import Units 
 
# ----------------------------------------------------------------------------------------------------------------------
#  Vertical Tail Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_vertical_tail_weight(vehicle, wing):
    """ Calculate the vertical tail weight

        Assumptions:
           Conventional tail configuration

        Source:
            The Flight Optimization System Weight Estimation Method

       Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.mass_properties.max_takeoff: MTOW                             [kilograms]
            wing - data dictionary with vertical tail properties                [dimensionless]
                -.taper: taper of wing
                -.areas.reference: surface area of wing                         [m^2]

       Outputs:
            WVT - vertical tail weight                                          [kilograms]

        Properties Used:
            N/A
        """
    DG          = vehicle.mass_properties.max_takeoff / Units.lbs  # Design gross weight in lb
    TRVT        = wing.taper
    NVERT       = 1  # Number of vertical tails
    WVT         = 0.32 * DG ** 0.3 * (TRVT + 0.5) * NVERT ** 0.7 * (wing.areas.reference/Units.ft**2)**0.85
    return WVT * Units.lbs

def tail_horizontal_FLOPS(vehicle, wing):
    """ Calculate the horizontal tail weight

        Assumptions:
           Conventional tail configuration

        Source:
            The Flight Optimization System Weight Estimation Method

       Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.mass_properties.max_takeoff: MTOW                             [kilograms]
            wing - data dictionary with vertical tail properties                [dimensionless]
                -.taper: taper of wing
                -.areas.reference: surface area of wing                         [m^2]

       Outputs:
            WHT - vertical tail weight                                          [kilograms]

        Properties Used:
            N/A
        """
    SHT     = wing.areas.reference / Units.ft **2
    DG      = vehicle.mass_properties.max_takeoff / Units.lbs
    TRHT    = wing.taper
    WHT     = 0.53 * SHT * DG ** 0.2 * (TRHT + 0.5)
    return WHT * Units.lbs




# RCAIDE/Library/Methods/Propulsors/Electric_Rotor_Propulsor/unpack_ice_propeller_unknowns.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke   

# ---------------------------------------------------------------------------------------------------------------------- 
#  unpack ice propeller network unknowns 
# ----------------------------------------------------------------------------------------------------------------------  

def unpack_ice_propeller_unknowns(propulsor,segment):  
    engine            = propulsor.engine 
    segment.state.conditions.energy[propulsor.tag][engine.tag].rpm = segment.state.unknowns[propulsor.tag + '_propeller_rpm'] 
    return 


# RCAIDE/Energy/Networks/Fuel.py
# 
# Created:  Oct 2023, M. Clarke
# Modified: 

# ----------------------------------------------------------------------------------------------------------------------
#  Imports
# ---------------------------------------------------------------------------------------------------------------------
# RCAIDE Imports
import  RCAIDE 
from RCAIDE.Framework.Mission.Common                      import Residuals 
from RCAIDE.Library.Mission.Common.Unpack_Unknowns.energy import unknowns
from .Network                                             import Network   

# ----------------------------------------------------------------------------------------------------------------------
# Fuel
# ----------------------------------------------------------------------------------------------------------------------  
class Fuel(Network):
    """ This is a  fuel-based network. 
    
        Assumptions:
        None
        
        Source: 
    """      
    
    def __defaults__(self):
        """ This sets the default values for the network to function.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            N/A
        """           

        self.tag                          = 'fuel'
        self.reverse_thrust               = False
        self.wing_mounted                 = True   
    # linking the different network components
    def evaluate(network,state,center_of_gravity):
        """ Calculate thrust given the current state of the vehicle
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs: 
    
            Outputs:  
        """           

        # Step 1: Unpack
        conditions     = state.conditions  
        fuel_lines     = network.fuel_lines 
        reverse_thrust = network.reverse_thrust
        total_thrust   = 0. * state.ones_row(3) 
        total_moment   = 0. * state.ones_row(3) 
        total_power    = 0. * state.ones_row(1) 
        total_mdot     = 0. * state.ones_row(1)   
        
        # Step 2: loop through compoments of network and determine performance
        for fuel_line in fuel_lines:     
            stored_results_flag  = False
            stored_propulsor_tag = None 
            for propulsor_group in fuel_line.assigned_propulsors:
                for propulsor_tag in propulsor_group:
                    propulsor =  network.propulsors[propulsor_tag]
                    if propulsor.active and fuel_line.active:   
                        if network.identical_propulsors == False:
                            # run analysis  
                            T,M,P,stored_results_flag,stored_propulsor_tag = propulsor.compute_performance(state,center_of_gravity)
                        else:             
                            if stored_results_flag == False: 
                                # run propulsor analysis 
                                T,M,P,stored_results_flag,stored_propulsor_tag = propulsor.compute_performance(state,center_of_gravity)
                            else:
                                # use previous propulsor results 
                                T,M,P = propulsor.reuse_stored_data(state,network,stored_propulsor_tag,center_of_gravity)
                          
                        total_thrust += T   
                        total_moment += M   
                        total_power  += P  
                
            # Step 2.2: Link each propulsor the its respective fuel tank(s)
            for fuel_tank in fuel_line.fuel_tanks:
                mdot = 0. * state.ones_row(1)   
                for propulsor in network.propulsors:
                    for source in (propulsor.active_fuel_tanks):
                        if fuel_tank.tag == source: 
                            mdot += conditions.energy[propulsor.tag].fuel_flow_rate 
                    
                # Step 2.3 : Determine cumulative fuel flow from fuel tank 
                fuel_tank_mdot = fuel_tank.fuel_selector_ratio*mdot + fuel_tank.secondary_fuel_flow 
                
                # Step 2.4: Store mass flow results 
                conditions.energy[fuel_line.tag][fuel_tank.tag].mass_flow_rate  = fuel_tank_mdot  
                total_mdot += fuel_tank_mdot                    
                            
        # Step 3: Pack results
        if reverse_thrust ==  True:
            total_thrust =  total_thrust* -1
            total_moment =  total_moment* -1
            
        conditions.energy.thrust_force_vector  = total_thrust
        conditions.energy.thrust_moment_vector = total_moment
        conditions.energy.power                = total_power 
        conditions.energy.vehicle_mass_rate    = total_mdot    
        
        return
    
    def unpack_unknowns(self,segment):
        """Unpacks the unknowns set in the mission to be available for the mission.

        Assumptions:
        N/A
        
        Source:
        N/A
        
        Inputs: 
            segment   - data structure of mission segment [-]
        
        Outputs: 
        
        Properties Used:
        N/A
        """            
         
        unknowns(segment)
        
        if issubclass(type(segment), type(RCAIDE.Framework.Mission.Segments.Ground)):
            pass 
        for network in segment.analyses.energy.vehicle.networks:
            for fuel_line_i, fuel_line in enumerate(network.fuel_lines):    
                if fuel_line.active:
                    for propulsor_group in  fuel_line.assigned_propulsors:
                        propulsor = network.propulsors[propulsor_group[0]]
                        propulsor.unpack_propulsor_unknowns(segment) 
        return    
     
    def residuals(self,segment):
        """ This packs the residuals to be sent to the mission solver.
   
           Assumptions:
           None
   
           Source:
           N/A
   
           Inputs:
           state.conditions.energy:
               motor(s).torque                      [N-m]
               rotor(s).torque                      [N-m] 
           residuals soecific to the battery cell   
           
           Outputs:
           residuals specific to battery cell and network
   
           Properties Used: 
           N/A
       """           
 
        for network in segment.analyses.energy.vehicle.networks:
            for fuel_line_i, fuel_line in enumerate(network.fuel_lines):    
                if fuel_line.active:
                    for propulsor_group in  fuel_line.assigned_propulsors:
                        propulsor =  network.propulsors[propulsor_group[0]]
                        propulsor.pack_propulsor_residuals(segment)  
        return      
    
    def add_unknowns_and_residuals_to_segment(self, segment):
        """ This function sets up the information that the mission needs to run a mission segment using this network 
         
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            segment
            eestimated_throttles           [-]
            estimated_propulsor_group_rpms [-]  
            
            Outputs:
            segment
    
            Properties Used:
            N/A
        """                   
        segment.state.residuals.network = Residuals()
        
        for network in segment.analyses.energy.vehicle.networks:
            for p_i, propulsor in enumerate(network.propulsors): 
                propulsor.append_operating_conditions(segment)           
    
                for tag, propulsor_item in  propulsor.items():  
                    if issubclass(type(propulsor_item), RCAIDE.Library.Components.Component):
                        propulsor_item.append_operating_conditions(segment,propulsor)            
    
            for fuel_line_i, fuel_line in enumerate(network.fuel_lines):   
                # ------------------------------------------------------------------------------------------------------            
                # Create fuel_line results data structure  
                # ------------------------------------------------------------------------------------------------------
                segment.state.conditions.energy[fuel_line.tag] = RCAIDE.Framework.Mission.Common.Conditions() 
                segment.state.conditions.noise[fuel_line.tag]  = RCAIDE.Framework.Mission.Common.Conditions()   
    
                # ------------------------------------------------------------------------------------------------------
                # Assign network-specific  residuals, unknowns and results data structures
                # ------------------------------------------------------------------------------------------------------
                if fuel_line.active:
                    for propulsor_group in  fuel_line.assigned_propulsors:
                        propulsor =  network.propulsors[propulsor_group[0]]
                        propulsor.append_propulsor_unknowns_and_residuals(segment)
    
                # ------------------------------------------------------------------------------------------------------
                # Assign sub component results data structures
                # ------------------------------------------------------------------------------------------------------  
                for fuel_tank in  fuel_line.fuel_tanks: 
                    fuel_tank.append_operating_conditions(segment,fuel_line)  
    
                for tag, fuel_line_item in  fuel_line.items():  
                    if issubclass(type(fuel_line_item), RCAIDE.Library.Components.Component):
                        fuel_line_item.append_operating_conditions(segment,fuel_line)                     
    
            for coolant_line_i, coolant_line in enumerate(network.coolant_lines):  
                # ------------------------------------------------------------------------------------------------------            
                # Create coolant_lines results data structure  
                # ------------------------------------------------------------------------------------------------------
                segment.state.conditions.energy[coolant_line.tag] = RCAIDE.Framework.Mission.Common.Conditions()        
                # ------------------------------------------------------------------------------------------------------
                # Assign network-specific  residuals, unknowns and results data structures
                # ------------------------------------------------------------------------------------------------------ 
    
                for battery_module in coolant_line.battery_modules: 
                    for btms in battery_module:
                        btms.append_operating_conditions(segment,coolant_line)
    
                for heat_exchanger in coolant_line.heat_exchangers: 
                    heat_exchanger.append_operating_conditions(segment, coolant_line)
    
                for reservoir in coolant_line.reservoirs: 
                    reservoir.append_operating_conditions(segment, coolant_line)
                     
        segment.process.iterate.unknowns.network   = self.unpack_unknowns      
        segment.process.iterate.residuals.network  = self.residuals
        
        return segment

    __call__ = evaluate     


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Transport/compute_horizontal_tail_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core    import Units
import numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Horizontal Tail Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_horizontal_tail_weight(vehicle, wing):
    """ Calculate the weight of the horizontal tail in a standard configuration
    
    Assumptions:
        calculated weight of the horizontal tail including the elevator
        Assume that the elevator is 25% of the horizontal tail 
    
    Source: 
        Aircraft Design: A Conceptual Approach by Raymer
        
    Inputs:
        wing.spans.projected - span of the horizontal tail                                              [meters]
        wing.sweeps.quarter_chord - sweep of the horizontal tail                                        [radians]
        Nult - ultimate design load of the aircraft                                                     [dimensionless]
        S_h - area of the horizontal tail                                                               [meters**2]
        vehicle.mass_properties.max_takeoff - maximum takeoff weight of the aircraft                    [kilograms]
        vehicle.wings['main_wing'].origin[0] - mean aerodynamic chord of the wing                       [meters]
        wing.aerodynamic_center[0]  - mean aerodynamic chord of the horizontal tail                     [meters]
        wing.thickness_to_chord  - thickness-to-chord ratio of the horizontal tail                      [dimensionless]
        wing.areas.exposed - exposed surface area for the horizontal tail                               [m^2]
        wing.areas.wetted - wetted surface area of tail
    
    Outputs:
        weight - weight of the horizontal tail                                                          [kilograms]
       
    Properties Used:
        N/A
    """
    # unpack inputs
    span       = wing.spans.projected / Units.ft  # Convert meters to ft
    sweep      = wing.sweeps.quarter_chord
    area       = wing.areas.reference / Units.ft ** 2  # Convert meters squared to ft squared
    mtow       = vehicle.mass_properties.max_takeoff / Units.lb  # Convert kg to lbs
    exposed    = wing.areas.exposed / wing.areas.wetted
    
    # Compute length between the main wing's aerodynamic center and the horizontal tail
    l_w2h      = np.array([wing.origin[0][0] + wing.aerodynamic_center[0] - vehicle.wings['main_wing'].origin[0][0] -  vehicle.wings['main_wing'].aerodynamic_center[0]])
    l_w        = np.array([vehicle.wings['main_wing'].chords.mean_aerodynamic / Units.ft])   # Convert from meters to ft
    length_w_h = l_w2h / Units.ft  # Distance from mean aerodynamic center of wing to mean aerodynamic center of
     
    # Calculate weight of wing for traditional aircraft horizontal tail
    weight_English = 5.25 * area + 0.8 * 10. ** -6 * vehicle.flight_envelope.ultimate_load * span ** 3. * mtow * l_w *\
                     np.sqrt(exposed * area) / (wing.thickness_to_chord * (np.cos(sweep) ** 2.) * length_w_h * area ** 1.5)

    weight = weight_English * Units.lbs  # Convert from lbs to kg

    return weight
